"""NumPy array quantity support for pintrs.

Provides ArrayQuantity that wraps numpy arrays with units, implementing
__array_ufunc__ for transparent numpy integration.
"""

# pyright: reportArgumentType=false, reportAttributeAccessIssue=false, reportInvalidTypeForm=false, reportPrivateUsage=false, reportRedeclaration=false, reportReturnType=false, reportUnknownArgumentType=false, reportUnknownMemberType=false, reportUnknownParameterType=false, reportUnknownVariableType=false, reportUnnecessaryIsInstance=false

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Iterator

    import numpy as np
    from numpy.typing import DTypeLike, NDArray

    from pintrs._core import Quantity as ScalarQuantity
    from pintrs._core import Unit
    from pintrs._core import UnitRegistry as _UnitRegistry

try:
    import numpy as np

    has_numpy = True
except ImportError:
    has_numpy = False

if has_numpy:
    from pintrs._core import Quantity as ScalarQuantity
    from pintrs._core import UnitRegistry as _UnitRegistry

    class ArrayQuantity:
        """A Quantity whose magnitude is a numpy array."""

        __hash__ = None  # type: ignore[assignment]  # mutable, unhashable

        def __init__(
            self,
            magnitude: NDArray[Any] | list[float],
            units: str,
            registry: _UnitRegistry | None = None,
        ) -> None:
            self._magnitude: NDArray[Any] = np.asarray(magnitude)
            self._units_str = units
            self._registry: _UnitRegistry = registry or _UnitRegistry()
            self._unit_obj: Unit = self._registry.Unit(units)

        @property
        def magnitude(self) -> NDArray[Any]:
            return self._magnitude

        @property
        def m(self) -> NDArray[Any]:
            return self._magnitude

        @property
        def units(self) -> Unit:
            return self._unit_obj

        @property
        def units_str(self) -> str:
            return self._units_str

        @property
        def u(self) -> Unit:
            return self._unit_obj

        @property
        def dimensionality(self) -> str:
            return str(self._unit_obj.dimensionality)

        @property
        def dimensionless(self) -> bool:
            return bool(self._unit_obj.dimensionless)

        @property
        def unitless(self) -> bool:
            return bool(self._unit_obj.dimensionless)

        @property
        def shape(self) -> tuple[int, ...]:
            result: tuple[int, ...] = self._magnitude.shape
            return result

        @property
        def ndim(self) -> int:
            return int(self._magnitude.ndim)

        @property
        def dtype(self) -> np.dtype[Any]:
            return self._magnitude.dtype  # type: ignore[no-any-return]

        @property
        def T(self) -> ArrayQuantity:  # noqa: N802
            return ArrayQuantity(
                self._magnitude.T,
                self._units_str,
                self._registry,
            )

        @property
        def flat(self) -> np.flatiter[Any]:
            return self._magnitude.flat  # type: ignore[no-any-return]

        @property
        def real(self) -> ArrayQuantity:
            return ArrayQuantity(
                np.real(self._magnitude),
                self._units_str,
                self._registry,
            )

        @property
        def imag(self) -> ArrayQuantity:
            return ArrayQuantity(
                np.imag(self._magnitude),
                self._units_str,
                self._registry,
            )

        def m_as(self, units: str) -> NDArray[Any]:
            factor = self._registry._get_conversion_factor(
                self._units_str,
                units,
            )
            return self._magnitude * factor

        def to(self, units: str) -> ArrayQuantity:
            factor = self._registry._get_conversion_factor(
                self._units_str,
                units,
            )
            return ArrayQuantity(
                self._magnitude * factor,
                units,
                self._registry,
            )

        def ito(self, units: str) -> None:
            factor = self._registry._get_conversion_factor(
                self._units_str,
                units,
            )
            self._magnitude = self._magnitude * factor
            self._units_str = units
            self._unit_obj = self._registry.Unit(units)

        def to_base_units(self) -> ArrayQuantity:
            factor, unit_str = self._registry._get_root_units(
                self._units_str,
            )
            return ArrayQuantity(
                self._magnitude * factor,
                unit_str,
                self._registry,
            )

        def to_root_units(self) -> ArrayQuantity:
            return self.to_base_units()

        def to_compact(self) -> ArrayQuantity:
            scalar_q = self._registry.Quantity(
                float(np.mean(np.abs(self._magnitude))),
                self._units_str,
            )
            compact = scalar_q.to_compact()
            target = str(compact.units)
            return self.to(target)

        def is_compatible_with(
            self,
            other: str | ArrayQuantity | ScalarQuantity,
        ) -> bool:
            scalar = self._registry.Quantity(1.0, self._units_str)
            if isinstance(other, str):
                return bool(scalar.is_compatible_with(other))
            if isinstance(other, ArrayQuantity):
                return bool(scalar.is_compatible_with(other._units_str))
            return bool(scalar.is_compatible_with(other))

        def check(self, dimension: str) -> bool:
            return bool(
                self._registry.Quantity(1.0, self._units_str).check(
                    dimension,
                ),
            )

        def clip(
            self,
            min: float | None = None,  # noqa: A002
            max: float | None = None,  # noqa: A002
        ) -> ArrayQuantity:
            return ArrayQuantity(
                self._magnitude.clip(min, max),
                self._units_str,
                self._registry,
            )

        def prod(self, axis: int | None = None) -> Any:
            return self._magnitude.prod(axis=axis)

        def tolist(self) -> list[float]:
            return list(self._magnitude.tolist())

        def fill(
            self,
            value: float | ScalarQuantity | ArrayQuantity,
        ) -> None:
            if isinstance(value, (ScalarQuantity, ArrayQuantity)):
                self._magnitude.fill(value.magnitude)
            else:
                self._magnitude.fill(value)

        def put(
            self,
            indices: NDArray[np.intp],
            values: float | ScalarQuantity | ArrayQuantity,
        ) -> None:
            if isinstance(values, (ScalarQuantity, ArrayQuantity)):
                self._magnitude.put(indices, values.magnitude)
            else:
                self._magnitude.put(indices, values)

        def searchsorted(
            self,
            v: float | ScalarQuantity | ArrayQuantity,
            side: str = "left",
        ) -> Any:
            mag = v.magnitude if isinstance(v, (ScalarQuantity, ArrayQuantity)) else v
            return self._magnitude.searchsorted(mag, side=side)

        def dot(
            self,
            other: ArrayQuantity | NDArray[Any],
        ) -> Any:
            mag = other._magnitude if isinstance(other, ArrayQuantity) else other
            return self._magnitude.dot(mag)

        def __len__(self) -> int:
            return len(self._magnitude)

        def __getitem__(
            self,
            key: int | slice | NDArray[np.intp],
        ) -> ArrayQuantity | ScalarQuantity:
            result = self._magnitude[key]
            if isinstance(result, np.ndarray):
                return ArrayQuantity(
                    result,
                    self._units_str,
                    self._registry,
                )
            return self._registry.Quantity(float(result), self._units_str)

        def __setitem__(
            self,
            key: int | slice | NDArray[np.intp],
            value: float | ScalarQuantity | ArrayQuantity,
        ) -> None:
            if isinstance(value, (ScalarQuantity, ArrayQuantity)):
                self._magnitude[key] = value.magnitude
            else:
                self._magnitude[key] = value

        def __iter__(self) -> Iterator[ArrayQuantity | ScalarQuantity]:
            for val in self._magnitude:
                if isinstance(val, np.ndarray):
                    yield ArrayQuantity(
                        val,
                        self._units_str,
                        self._registry,
                    )
                else:
                    yield self._registry.Quantity(
                        float(val),
                        self._units_str,
                    )

        def __repr__(self) -> str:
            return f"<Quantity({self._magnitude}, '{self._units_str}')>"

        def __str__(self) -> str:
            return f"{self._magnitude} {self._units_str}"

        def __neg__(self) -> ArrayQuantity:
            return ArrayQuantity(
                -self._magnitude,
                self._units_str,
                self._registry,
            )

        def __abs__(self) -> ArrayQuantity:
            return ArrayQuantity(
                np.abs(self._magnitude),
                self._units_str,
                self._registry,
            )

        def _get_other_mag_and_units(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> tuple[NDArray[Any] | float, str | None]:
            if isinstance(other, ArrayQuantity):
                return other._magnitude, other._units_str
            if isinstance(other, ScalarQuantity):
                return other.magnitude, str(other.units)
            return other, None

        def __add__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None and units != self._units_str:
                factor = self._registry._get_conversion_factor(
                    units,
                    self._units_str,
                )
                mag = mag * factor
            return ArrayQuantity(
                self._magnitude + mag,
                self._units_str,
                self._registry,
            )

        def __radd__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            return self.__add__(other)

        def __sub__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None and units != self._units_str:
                factor = self._registry._get_conversion_factor(
                    units,
                    self._units_str,
                )
                mag = mag * factor
            return ArrayQuantity(
                self._magnitude - mag,
                self._units_str,
                self._registry,
            )

        def __rsub__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            result = self.__sub__(other)
            return ArrayQuantity(
                -result._magnitude,
                result._units_str,
                result._registry,
            )

        def __mul__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None:
                new_units = str(
                    self._unit_obj * self._registry.Unit(units),
                )
            else:
                new_units = self._units_str
            return ArrayQuantity(
                self._magnitude * mag,
                new_units,
                self._registry,
            )

        def __rmul__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            return self.__mul__(other)

        def __truediv__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None:
                new_units = str(
                    self._unit_obj / self._registry.Unit(units),
                )
            else:
                new_units = self._units_str
            return ArrayQuantity(
                self._magnitude / mag,
                new_units,
                self._registry,
            )

        def __rtruediv__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> ArrayQuantity:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None:
                new_units = str(
                    self._registry.Unit(units) / self._unit_obj,
                )
            else:
                new_units = str(
                    self._registry.Unit("dimensionless") / self._unit_obj,
                )
            return ArrayQuantity(
                mag / self._magnitude,
                new_units,
                self._registry,
            )

        def __pow__(self, exp: float) -> ArrayQuantity:
            new_units = str(self._unit_obj**exp)
            return ArrayQuantity(
                self._magnitude**exp,
                new_units,
                self._registry,
            )

        def __eq__(self, other: object) -> bool:
            if isinstance(other, ArrayQuantity):
                if other._units_str != self._units_str:
                    factor = self._registry._get_conversion_factor(
                        other._units_str,
                        self._units_str,
                    )
                    return bool(
                        np.array_equal(self._magnitude, other._magnitude * factor),
                    )
                return bool(np.array_equal(self._magnitude, other._magnitude))
            if isinstance(other, ScalarQuantity):
                return bool(np.array_equal(self._magnitude, other.magnitude))
            if isinstance(other, (np.ndarray, list, tuple, int, float, complex)):
                return bool(np.array_equal(self._magnitude, other))
            return False

        def __ne__(self, other: object) -> bool:
            return not self.__eq__(other)

        def __lt__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> Any:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None and units != self._units_str:
                factor = self._registry._get_conversion_factor(
                    units,
                    self._units_str,
                )
                mag = mag * factor
            return self._magnitude < mag

        def __le__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> Any:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None and units != self._units_str:
                factor = self._registry._get_conversion_factor(
                    units,
                    self._units_str,
                )
                mag = mag * factor
            return self._magnitude <= mag

        def __gt__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> Any:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None and units != self._units_str:
                factor = self._registry._get_conversion_factor(
                    units,
                    self._units_str,
                )
                mag = mag * factor
            return self._magnitude > mag

        def __ge__(
            self,
            other: ArrayQuantity | ScalarQuantity | float,
        ) -> Any:
            mag, units = self._get_other_mag_and_units(other)
            if units is not None and units != self._units_str:
                factor = self._registry._get_conversion_factor(
                    units,
                    self._units_str,
                )
                mag = mag * factor
            return self._magnitude >= mag

        def __bool__(self) -> bool:
            return bool(np.any(self._magnitude != 0))

        def __array__(
            self,
            dtype: DTypeLike | None = None,
        ) -> NDArray[Any]:
            if dtype is not None:
                return np.asarray(self._magnitude, dtype=dtype)
            return np.asarray(self._magnitude)

        # Ufuncs that require dimensionless input
        _DIMENSIONLESS_UFUNCS: frozenset[np.ufunc] = frozenset()
        # Trig ufuncs that accept radians/degrees/dimensionless
        _TRIG_UFUNCS: frozenset[np.ufunc] = frozenset()
        # Inverse trig ufuncs that return radians
        _INVERSE_TRIG_UFUNCS: frozenset[np.ufunc] = frozenset()
        # arctan2: two args with matching units, returns radians
        _ARCTAN2_UFUNCS: frozenset[np.ufunc] = frozenset()
        # Ufuncs that require matching dimensions for two inputs
        _MATCHING_DIMS_UFUNCS: frozenset[np.ufunc] = frozenset()

        @staticmethod
        def _ufunc_output_units(
            ufunc: np.ufunc,
            input_units: list[Unit | None],
            processed: list[NDArray[Any] | float],
            default_units: str,
        ) -> str:
            """Compute output units for a numpy ufunc."""
            if ufunc in (np.multiply, np.matmul):
                u0 = input_units[0] if input_units else None
                u1 = input_units[1] if len(input_units) > 1 else None
                if u0 is not None and u1 is not None:
                    return str(u0 * u1)
                if u0 is not None:
                    return str(u0)
                return str(u1) if u1 is not None else default_units
            if ufunc in (np.divide, np.true_divide, np.floor_divide):
                u0 = input_units[0] if input_units else None
                u1 = input_units[1] if len(input_units) > 1 else None
                if u0 is not None and u1 is not None:
                    return str(u0 / u1)
                return str(u0) if u0 is not None else default_units
            if ufunc is np.power:
                exp = processed[1] if len(processed) > 1 else 1
                if isinstance(exp, np.ndarray):
                    exp = float(exp.flat[0])
                u0 = input_units[0]
                return str(u0 ** float(exp)) if u0 is not None else default_units
            power_map: dict[np.ufunc, float] = {
                np.sqrt: 0.5,
                np.square: 2.0,
                np.cbrt: 1.0 / 3.0,
                np.reciprocal: -1.0,
            }
            if ufunc in power_map:
                u0 = input_units[0]
                return str(u0 ** power_map[ufunc]) if u0 is not None else default_units
            return default_units

        def _check_additive_compat(
            self,
            ufunc: np.ufunc,
            inputs: tuple[Any, ...],
            processed: list[NDArray[Any] | float],
            input_units: list[Unit | None],
        ) -> list[NDArray[Any] | float]:
            """Check dimensionality and convert units for add/subtract."""
            from pintrs import DimensionalityError  # noqa: PLC0415

            if ufunc not in self._MATCHING_DIMS_UFUNCS:
                return processed
            units_with_idx = [
                (i, u) for i, u in enumerate(input_units) if u is not None
            ]
            if len(units_with_idx) < 2:
                return processed
            _, base_unit = units_with_idx[0]
            result = list(processed)
            for idx, unit in units_with_idx[1:]:
                if str(unit.dimensionality) != str(base_unit.dimensionality):
                    msg = f"Cannot convert from '{unit}' to '{base_unit}'"
                    raise DimensionalityError(msg)
                if str(unit) != str(base_unit):
                    inp = inputs[idx]
                    if isinstance(inp, ArrayQuantity):
                        converted = inp.to(str(base_unit))
                        result[idx] = converted._magnitude
                    elif isinstance(inp, ScalarQuantity):
                        converted_q = inp.to(str(base_unit))
                        result[idx] = converted_q.magnitude
            return result

        def _check_trig_units(
            self,
            ufunc: np.ufunc,
            inputs: tuple[Any, ...],
            processed: list[NDArray[Any] | float],
            input_units: list[Unit | None],
        ) -> list[NDArray[Any] | float]:
            """Validate and convert units for trig/exp/log ufuncs."""
            from pintrs import DimensionalityError  # noqa: PLC0415

            if ufunc in self._TRIG_UFUNCS:
                u = input_units[0] if input_units else None
                if u is not None:
                    u_str = str(u)
                    if u_str in ("rad", "radian"):
                        return processed
                    if u_str in ("deg", "degree"):
                        result = list(processed)
                        result[0] = np.deg2rad(processed[0])
                        return result
                    q_check = self._registry.Quantity(1.0, u_str)
                    if q_check.dimensionless:
                        return processed
                    msg = (
                        f"Cannot apply '{ufunc.__name__}' to quantity "
                        f"with units '{u_str}'"
                    )
                    raise DimensionalityError(msg)
            if ufunc in self._INVERSE_TRIG_UFUNCS:
                u = input_units[0] if input_units else None
                if u is not None:
                    q_check = self._registry.Quantity(1.0, str(u))
                    if not q_check.dimensionless:
                        msg = (
                            f"Cannot apply '{ufunc.__name__}' to quantity "
                            f"with units '{u}'"
                        )
                        raise DimensionalityError(msg)
            if ufunc in self._DIMENSIONLESS_UFUNCS:
                u = input_units[0] if input_units else None
                if u is not None:
                    q_check = self._registry.Quantity(1.0, str(u))
                    if not q_check.dimensionless:
                        msg = (
                            f"Cannot apply '{ufunc.__name__}' to quantity "
                            f"with units '{u}'"
                        )
                        raise DimensionalityError(msg)
            return processed

        _COMPARISON_UFUNCS: frozenset[np.ufunc] = frozenset()

        def __array_ufunc__(
            self,
            ufunc: np.ufunc,
            method: str,
            *inputs: ArrayQuantity | ScalarQuantity | float,
            **kwargs: object,
        ) -> Any:
            """Support numpy ufuncs with unit tracking."""
            processed: list[NDArray[Any] | float] = []
            input_units: list[Unit | None] = []
            for inp in inputs:
                if isinstance(inp, ArrayQuantity):
                    processed.append(inp._magnitude)
                    input_units.append(inp._unit_obj)
                elif isinstance(inp, ScalarQuantity):
                    processed.append(inp.magnitude)
                    input_units.append(inp.units)
                else:
                    processed.append(inp)
                    input_units.append(None)

            processed = self._check_additive_compat(
                ufunc,
                inputs,
                processed,
                input_units,
            )
            processed = self._check_trig_units(
                ufunc,
                inputs,
                processed,
                input_units,
            )

            result = getattr(ufunc, method)(*processed, **kwargs)

            if ufunc in self._COMPARISON_UFUNCS:
                return result

            if ufunc in self._TRIG_UFUNCS or ufunc in self._DIMENSIONLESS_UFUNCS:
                out_units = "dimensionless"
            elif ufunc in self._INVERSE_TRIG_UFUNCS or ufunc in self._ARCTAN2_UFUNCS:
                out_units = "radian"
            else:
                out_units = self._ufunc_output_units(
                    ufunc,
                    input_units,
                    processed,
                    self._units_str,
                )

            if isinstance(result, np.ndarray):
                return ArrayQuantity(result, out_units, self._registry)
            if np.isscalar(result):
                magnitude = result.item() if isinstance(result, np.generic) else result
                if isinstance(magnitude, (int, float)):
                    return self._registry.Quantity(float(magnitude), out_units)
                if isinstance(magnitude, str):
                    return self._registry.Quantity(magnitude, out_units)
            return result

        def __array_function__(
            self,
            func: Any,
            types: tuple[type[Any], ...],
            args: tuple[Any, ...],
            kwargs: dict[str, Any],
        ) -> Any:
            if not all(issubclass(t, ArrayQuantity) for t in types):
                return NotImplemented

            def _scalar(val: Any, units: str) -> Any:
                return self._registry.Quantity(float(val), units)

            def _array(val: Any, units: str) -> ArrayQuantity:
                return ArrayQuantity(np.asarray(val), units, self._registry)

            if func is np.sum:
                result = np.sum(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.mean:
                result = np.mean(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.min or func is np.amin:
                result = np.min(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.max or func is np.amax:
                result = np.max(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.std:
                result = np.std(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.var:
                result = np.var(self._magnitude, **kwargs)
                u = self._unit_obj
                return _scalar(result, str(u * u))
            if func is np.cumsum:
                result = np.cumsum(self._magnitude, **kwargs)
                return _array(result, self._units_str)
            if func is np.cumprod:
                result = np.cumprod(self._magnitude, **kwargs)
                return _array(result, self._units_str)
            if func is np.argmin:
                return np.argmin(self._magnitude, **kwargs)
            if func is np.argmax:
                return np.argmax(self._magnitude, **kwargs)
            if func is np.sort:
                result = np.sort(self._magnitude, **kwargs)
                return _array(result, self._units_str)
            if func is np.argsort:
                return np.argsort(self._magnitude, **kwargs)
            if func is np.isreal:
                return np.isreal(self._magnitude)
            if func is np.iscomplex:
                return np.iscomplex(self._magnitude)
            if func is np.reshape:
                new_shape = (
                    args[1]
                    if len(args) > 1
                    else kwargs.get("newshape", kwargs.get("shape"))
                )
                result = np.reshape(self._magnitude, new_shape)  # pyright: ignore[reportCallIssue]
                return _array(result, self._units_str)
            if func is np.ones_like:
                result = np.ones_like(self._magnitude, **kwargs)
                return _array(result, self._units_str)
            if func is np.zeros_like:
                result = np.zeros_like(self._magnitude, **kwargs)
                return _array(result, self._units_str)
            if func is np.full_like:
                fill = args[1] if len(args) > 1 else kwargs.get("fill_value", 0)
                if isinstance(fill, (ArrayQuantity, ScalarQuantity)):
                    fill = fill.magnitude
                result = np.full_like(self._magnitude, fill, **kwargs)
                return _array(result, self._units_str)
            if func is np.concatenate:
                arrays = args[0]
                mags = [
                    a._magnitude if isinstance(a, ArrayQuantity) else a for a in arrays
                ]
                result = np.concatenate(mags, **kwargs)
                return _array(result, self._units_str)
            if func is np.stack:
                arrays = args[0]
                mags = [
                    a._magnitude if isinstance(a, ArrayQuantity) else a for a in arrays
                ]
                result = np.stack(mags, **kwargs)
                return _array(result, self._units_str)
            if func is np.where:
                condition = args[0]
                if len(args) >= 3:
                    x = (
                        args[1]._magnitude
                        if isinstance(args[1], ArrayQuantity)
                        else args[1]
                    )
                    y = (
                        args[2]._magnitude
                        if isinstance(args[2], ArrayQuantity)
                        else args[2]
                    )
                    result = np.where(condition, x, y)
                    return _array(result, self._units_str)
                return np.where(condition)
            if func is np.clip:
                arr = (
                    args[0]._magnitude
                    if isinstance(args[0], ArrayQuantity)
                    else args[0]
                )
                a_min = kwargs.get("a_min", args[1] if len(args) > 1 else None)
                a_max = kwargs.get("a_max", args[2] if len(args) > 2 else None)
                if isinstance(a_min, (ArrayQuantity, ScalarQuantity)):
                    a_min = a_min.magnitude  # pyright: ignore[reportOptionalMemberAccess]
                if isinstance(a_max, (ArrayQuantity, ScalarQuantity)):
                    a_max = a_max.magnitude  # pyright: ignore[reportOptionalMemberAccess]
                result = np.clip(arr, a_min, a_max)
                return _array(result, self._units_str)
            if func is np.ptp:
                result = np.ptp(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.median:
                result = np.median(self._magnitude, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.percentile:
                q_val = args[1] if len(args) > 1 else kwargs.get("q")
                result = np.percentile(self._magnitude, q_val, **kwargs)
                return _scalar(result, self._units_str)
            if func is np.dot:
                a, b = args[0], args[1]
                a_mag = a._magnitude if isinstance(a, ArrayQuantity) else np.asarray(a)
                b_mag = b._magnitude if isinstance(b, ArrayQuantity) else np.asarray(b)
                a_u = a._unit_obj if isinstance(a, ArrayQuantity) else None
                b_u = b._unit_obj if isinstance(b, ArrayQuantity) else None
                result = np.dot(a_mag, b_mag, **kwargs)
                if a_u is not None and b_u is not None:
                    out_u = str(a_u * b_u)
                elif a_u is not None:
                    out_u = str(a_u)
                elif b_u is not None:
                    out_u = str(b_u)
                else:
                    out_u = "dimensionless"
                if isinstance(result, np.ndarray):
                    return _array(result, out_u)
                return _scalar(result, out_u)
            if func is np.cross:
                a, b = args[0], args[1]
                a_mag = a._magnitude if isinstance(a, ArrayQuantity) else np.asarray(a)
                b_mag = b._magnitude if isinstance(b, ArrayQuantity) else np.asarray(b)
                a_u = a._unit_obj if isinstance(a, ArrayQuantity) else None
                b_u = b._unit_obj if isinstance(b, ArrayQuantity) else None
                result = np.cross(a_mag, b_mag, **kwargs)
                if a_u is not None and b_u is not None:
                    out_u = str(a_u * b_u)
                elif a_u is not None:
                    out_u = str(a_u)
                elif b_u is not None:
                    out_u = str(b_u)
                else:
                    out_u = "dimensionless"
                if isinstance(result, np.ndarray):
                    return _array(result, out_u)
                return _scalar(result, out_u)
            if func is np.outer:
                a, b = args[0], args[1]
                a_mag = a._magnitude if isinstance(a, ArrayQuantity) else np.asarray(a)
                b_mag = b._magnitude if isinstance(b, ArrayQuantity) else np.asarray(b)
                a_u = a._unit_obj if isinstance(a, ArrayQuantity) else None
                b_u = b._unit_obj if isinstance(b, ArrayQuantity) else None
                result = np.outer(a_mag, b_mag, **kwargs)
                if a_u is not None and b_u is not None:
                    out_u = str(a_u * b_u)
                elif a_u is not None:
                    out_u = str(a_u)
                elif b_u is not None:
                    out_u = str(b_u)
                else:
                    out_u = "dimensionless"
                return _array(result, out_u)
            if func is np.inner:
                a, b = args[0], args[1]
                a_mag = a._magnitude if isinstance(a, ArrayQuantity) else np.asarray(a)
                b_mag = b._magnitude if isinstance(b, ArrayQuantity) else np.asarray(b)
                a_u = a._unit_obj if isinstance(a, ArrayQuantity) else None
                b_u = b._unit_obj if isinstance(b, ArrayQuantity) else None
                result = np.inner(a_mag, b_mag, **kwargs)
                if a_u is not None and b_u is not None:
                    out_u = str(a_u * b_u)
                elif a_u is not None:
                    out_u = str(a_u)
                elif b_u is not None:
                    out_u = str(b_u)
                else:
                    out_u = "dimensionless"
                if isinstance(result, np.ndarray):
                    return _array(result, out_u)
                return _scalar(result, out_u)
            if func is np.vdot:
                a, b = args[0], args[1]
                a_mag = a._magnitude if isinstance(a, ArrayQuantity) else np.asarray(a)
                b_mag = b._magnitude if isinstance(b, ArrayQuantity) else np.asarray(b)
                a_u = a._unit_obj if isinstance(a, ArrayQuantity) else None
                b_u = b._unit_obj if isinstance(b, ArrayQuantity) else None
                result = np.vdot(a_mag, b_mag)
                if a_u is not None and b_u is not None:
                    out_u = str(a_u * b_u)
                elif a_u is not None:
                    out_u = str(a_u)
                elif b_u is not None:
                    out_u = str(b_u)
                else:
                    out_u = "dimensionless"
                return _scalar(result, out_u)
            if func is np.correlate:
                a, b = args[0], args[1]
                a_mag = a._magnitude if isinstance(a, ArrayQuantity) else np.asarray(a)
                b_mag = b._magnitude if isinstance(b, ArrayQuantity) else np.asarray(b)
                a_u = a._unit_obj if isinstance(a, ArrayQuantity) else None
                b_u = b._unit_obj if isinstance(b, ArrayQuantity) else None
                result = np.correlate(a_mag, b_mag, **kwargs)
                if a_u is not None and b_u is not None:
                    out_u = str(a_u * b_u)
                elif a_u is not None:
                    out_u = str(a_u)
                elif b_u is not None:
                    out_u = str(b_u)
                else:
                    out_u = "dimensionless"
                return _array(result, out_u)
            if func is np.trapezoid:
                y = args[0]
                y_mag = y._magnitude if isinstance(y, ArrayQuantity) else np.asarray(y)
                y_u = y._unit_obj if isinstance(y, ArrayQuantity) else None
                trap_kwargs = dict(kwargs)
                x = args[1] if len(args) > 1 else kwargs.get("x")
                if x is not None and "x" in trap_kwargs:
                    del trap_kwargs["x"]
                x_u = None
                if x is not None:
                    if isinstance(x, ArrayQuantity):
                        x_u = x._unit_obj
                        x_mag = x._magnitude
                    else:
                        x_mag = np.asarray(x)
                    result = np.trapezoid(y_mag, x_mag, **trap_kwargs)
                else:
                    result = np.trapezoid(y_mag, **trap_kwargs)
                if y_u is not None and x_u is not None:
                    out_u = str(y_u * x_u)
                elif y_u is not None:
                    out_u = str(y_u)
                elif x_u is not None:
                    out_u = str(x_u)
                else:
                    out_u = "dimensionless"
                if isinstance(result, np.ndarray):
                    return _array(result, out_u)
                return _scalar(result, out_u)
            return NotImplemented

    ArrayQuantity._COMPARISON_UFUNCS = frozenset(
        {
            np.equal,
            np.not_equal,
            np.less,
            np.less_equal,
            np.greater,
            np.greater_equal,
            np.isnan,
            np.isinf,
            np.isfinite,
            np.signbit,
            np.logical_and,
            np.logical_or,
            np.logical_not,
        },
    )

    ArrayQuantity._DIMENSIONLESS_UFUNCS = frozenset(
        {
            np.exp,
            np.exp2,
            np.expm1,
            np.log,
            np.log2,
            np.log10,
            np.log1p,
            np.logaddexp,
            np.logaddexp2,
        },
    )

    ArrayQuantity._TRIG_UFUNCS = frozenset(
        {
            np.sin,
            np.cos,
            np.tan,
            np.sinh,
            np.cosh,
            np.tanh,
        },
    )

    ArrayQuantity._INVERSE_TRIG_UFUNCS = frozenset(
        {
            np.arcsin,
            np.arccos,
            np.arctan,
            np.arcsinh,
            np.arccosh,
            np.arctanh,
        },
    )

    # arctan2 takes two args with matching units and returns radians
    ArrayQuantity._ARCTAN2_UFUNCS = frozenset({np.arctan2})

    ArrayQuantity._MATCHING_DIMS_UFUNCS = frozenset(
        {
            np.add,
            np.subtract,
            np.remainder,
            np.mod,
            np.fmod,
            np.copysign,
            np.nextafter,
            np.greater,
            np.greater_equal,
            np.less,
            np.less_equal,
            np.equal,
            np.not_equal,
            np.maximum,
            np.minimum,
            np.hypot,
            np.arctan2,
        },
    )

else:

    class _ArrayQuantityUnavailable:
        """Placeholder that fails on use when NumPy is not installed."""

        def __init__(self, *_args: object, **_kwargs: object) -> None:
            msg = (
                "NumPy support requires numpy. Install pintrs[numpy] or install numpy."
            )
            raise ModuleNotFoundError(msg)

    ArrayQuantity: Any = _ArrayQuantityUnavailable  # type: ignore[no-redef]



"""Compatibility/compat tests - ported from pint's test_compat.py.

Tests NaN handling, equality comparison, and zero detection behavior
through the quantity API rather than pint's internal compat functions.
"""

from __future__ import annotations

import math

import pytest
from pintrs import UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestQuantityNaN:
    """Test NaN behavior in quantities (mirrors pint's isnan tests)."""

    def test_nan_magnitude(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(float("nan"), "meter")
        assert math.isnan(q.magnitude)

    def test_nan_not_equal_to_itself(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(float("nan"), "meter")
        assert q != q

    def test_nan_not_equal_to_number(self, ureg: UnitRegistry) -> None:
        q_nan = ureg.Quantity(float("nan"), "meter")
        q_num = ureg.Quantity(1.0, "meter")
        assert q_nan != q_num

    def test_zero_is_not_nan(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.0, "meter")
        assert not math.isnan(q.magnitude)

    def test_regular_float_is_not_nan(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter")
        assert not math.isnan(q.magnitude)


class TestQuantityNaNNumpy:
    """Test NaN behavior with numpy arrays in quantities via ArrayQuantity."""

    def test_numpy_nan_in_array(self, ureg: UnitRegistry) -> None:
        np = pytest.importorskip("numpy")
        from pintrs import ArrayQuantity

        q = ArrayQuantity(np.array([1.0, np.nan, 3.0]), "meter", ureg)
        assert np.any(np.isnan(q.magnitude))

    def test_numpy_no_nan(self, ureg: UnitRegistry) -> None:
        np = pytest.importorskip("numpy")
        from pintrs import ArrayQuantity

        q = ArrayQuantity(np.array([1.0, 2.0, 3.0]), "meter", ureg)
        assert not np.any(np.isnan(q.magnitude))

    def test_numpy_all_nan(self, ureg: UnitRegistry) -> None:
        np = pytest.importorskip("numpy")
        from pintrs import ArrayQuantity

        q = ArrayQuantity(np.array([np.nan, np.nan]), "meter", ureg)
        assert np.all(np.isnan(q.magnitude))


class TestQuantityEquality:
    """Test equality comparison (mirrors pint's eq tests)."""

    def test_equal_same_units(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") == ureg.Quantity(1.0, "meter")

    def test_not_equal_different_magnitude(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") != ureg.Quantity(2.0, "meter")

    def test_equal_compatible_units(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "kilometer") == ureg.Quantity(1000.0, "meter")

    def test_not_equal_different_values(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") != ureg.Quantity(1.0, "second")


class TestQuantityEqualityNumpy:
    """Test equality with numpy arrays."""

    def test_array_equal(self, ureg: UnitRegistry) -> None:
        np = pytest.importorskip("numpy")
        from pintrs import ArrayQuantity

        a = ArrayQuantity(np.array([1.0, 2.0]), "meter", ureg)
        b = ArrayQuantity(np.array([1.0, 2.0]), "meter", ureg)
        np.testing.assert_array_equal(a.magnitude, b.magnitude)

    def test_array_not_equal(self, ureg: UnitRegistry) -> None:
        np = pytest.importorskip("numpy")
        from pintrs import ArrayQuantity

        a = ArrayQuantity(np.array([1.0, 2.0]), "meter", ureg)
        b = ArrayQuantity(np.array([1.0, 3.0]), "meter", ureg)
        assert not np.array_equal(a.magnitude, b.magnitude)


class TestQuantityZero:
    """Test zero detection (mirrors pint's zero_or_nan tests)."""

    def test_zero_is_falsy(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "meter")
        assert not bool(q)

    def test_nonzero_is_truthy(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "meter")
        assert bool(q)

    def test_zero_magnitude(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "meter")
        assert q.magnitude == 0

    def test_nan_is_truthy(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(float("nan"), "meter")
        assert bool(q)



"""Comprehensive Unit tests - ported from pint's test suite."""

from __future__ import annotations

import copy
import pickle

import pytest
from pintrs import (
    DimensionalityError,
    UndefinedUnitError,
    Unit,
    UnitRegistry,
)


@pytest.fixture
def ureg() -> UnitRegistry:
    """Fresh registry for each test."""
    return UnitRegistry()


# ---------------------------------------------------------------------------
# 1. TestUnitCreation
# ---------------------------------------------------------------------------


class TestUnitCreation:
    """Unit creation from strings, parse_units, and compound expressions."""

    def test_create_from_string(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        assert str(u) == "m"

    def test_create_from_symbol(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("m")
        assert str(u) == "m"

    def test_create_from_parse_units(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        assert isinstance(u, Unit)
        assert str(u) == "m"

    def test_create_compound_mul(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter * second")
        assert "meter" in dict(u._units_dict())
        assert "second" in dict(u._units_dict())

    def test_create_compound_div(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(1.0)
        assert d["second"] == pytest.approx(-1.0)

    def test_create_compound_pow(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2")
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(2.0)

    def test_create_complex_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("kilogram * meter / second ** 2")
        d = dict(u._units_dict())
        assert d["second"] == pytest.approx(-2.0)
        assert d["meter"] == pytest.approx(1.0)

    def test_create_dimensionless(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("dimensionless")
        assert u.dimensionless

    def test_create_prefixed(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("kilometer")
        assert str(u) in ("km", "kilometer")

    def test_create_from_abbreviation(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("kg")
        assert u.dimensionality == ureg.parse_units("kilogram").dimensionality

    def test_create_undefined_unit_raises(self, ureg: UnitRegistry) -> None:
        with pytest.raises(UndefinedUnitError):
            ureg.parse_units("nonexistent_xyzzy_unit")

    def test_unit_returns_unit_type(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        assert type(u).__name__ == "Unit"

    def test_parse_units_returns_unit_type(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        assert type(u).__name__ == "Unit"

    def test_getattr_returns_quantity(self, ureg: UnitRegistry) -> None:
        q = ureg.meter
        assert type(q).__name__ == "Quantity"
        assert q.magnitude == pytest.approx(1.0)


# ---------------------------------------------------------------------------
# 2. TestUnitArithmetic
# ---------------------------------------------------------------------------


class TestUnitArithmetic:
    """Multiplication, division, and exponentiation of units."""

    def test_mul_units(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") * ureg.parse_units("second")
        assert isinstance(u, Unit)

    def test_div_units(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") / ureg.parse_units("second")
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(1.0)
        assert d["second"] == pytest.approx(-1.0)

    def test_pow_unit(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") ** 3
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(3.0)

    def test_pow_zero_gives_dimensionless(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") ** 0
        assert u.dimensionless

    def test_pow_negative(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") ** -1
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(-1.0)

    def test_pow_fractional(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") ** 0.5
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(0.5)

    def test_mul_number_returns_quantity(self, ureg: UnitRegistry) -> None:
        result = ureg.parse_units("meter") * 3.0
        assert type(result).__name__ == "Quantity"
        assert result.magnitude == pytest.approx(3.0)

    def test_rmul_number_returns_quantity(self, ureg: UnitRegistry) -> None:
        result = 3.0 * ureg.parse_units("meter")
        assert type(result).__name__ == "Quantity"
        assert result.magnitude == pytest.approx(3.0)

    def test_compound_mul_then_div(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        s = ureg.parse_units("second")
        u = (m * m) / s
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(2.0)
        assert d["second"] == pytest.approx(-1.0)

    def test_mul_dimensionless_identity(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        d = ureg.parse_units("dimensionless")
        result = m * d
        assert result == m


# ---------------------------------------------------------------------------
# 3. TestUnitComparison
# ---------------------------------------------------------------------------


class TestUnitComparison:
    """Equality, hashing, and compatibility checks."""

    def test_eq_same_unit(self, ureg: UnitRegistry) -> None:
        assert ureg.parse_units("meter") == ureg.parse_units("meter")

    def test_eq_symbol_and_name(self, ureg: UnitRegistry) -> None:
        assert ureg.parse_units("m") == ureg.parse_units("meter")

    def test_ne_different_units(self, ureg: UnitRegistry) -> None:
        assert ureg.parse_units("meter") != ureg.parse_units("second")

    def test_ne_different_dimension(self, ureg: UnitRegistry) -> None:
        assert ureg.parse_units("meter") != ureg.parse_units("kilogram")

    def test_ne_different_prefix(self, ureg: UnitRegistry) -> None:
        assert ureg.parse_units("meter") != ureg.parse_units("kilometer")

    def test_eq_non_unit_returns_false(self, ureg: UnitRegistry) -> None:
        assert ureg.parse_units("meter") != 42
        assert ureg.parse_units("meter") != "meter"

    def test_hash_equal_units(self, ureg: UnitRegistry) -> None:
        u1 = ureg.parse_units("meter")
        u2 = ureg.parse_units("meter")
        assert hash(u1) == hash(u2)

    def test_hash_different_units(self, ureg: UnitRegistry) -> None:
        u1 = ureg.parse_units("meter")
        u2 = ureg.parse_units("second")
        assert hash(u1) != hash(u2)

    def test_unit_in_set(self, ureg: UnitRegistry) -> None:
        s = {ureg.parse_units("meter"), ureg.parse_units("meter")}
        assert len(s) == 1

    def test_unit_as_dict_key(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        d = {u: "length"}
        assert d[ureg.parse_units("meter")] == "length"

    def test_compatible_with_unit(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        km = ureg.parse_units("kilometer")
        assert m.is_compatible_with(km)

    def test_compatible_with_string(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        assert m.is_compatible_with("kilometer")

    def test_not_compatible_different_dimension(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        assert not m.is_compatible_with("second")

    def test_compatible_units_list(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        compat = m.compatible_units()
        assert isinstance(compat, list)
        assert len(compat) > 0
        assert "meter" in compat
        assert "foot" in compat

    def test_compound_units_equal(self, ureg: UnitRegistry) -> None:
        u1 = ureg.parse_units("meter / second")
        u2 = ureg.parse_units("meter") / ureg.parse_units("second")
        assert u1 == u2


# ---------------------------------------------------------------------------
# 4. TestUnitDimensionality
# ---------------------------------------------------------------------------


class TestUnitDimensionality:
    """Dimensionality strings and dimensionless property."""

    @pytest.mark.parametrize(
        ("unit_str", "expected_dim"),
        [
            ("meter", "[length]"),
            ("second", "[time]"),
            ("kilogram", "[mass]"),
            ("ampere", "[current]"),
            ("kelvin", "[temperature]"),
            ("mole", "[substance]"),
            ("candela", "[luminosity]"),
        ],
    )
    def test_base_unit_dimensionality(
        self,
        ureg: UnitRegistry,
        unit_str: str,
        expected_dim: str,
    ) -> None:
        u = ureg.parse_units(unit_str)
        assert u.dimensionality == expected_dim

    def test_compound_dimensionality(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        assert "[length]" in u.dimensionality
        assert "[time]" in u.dimensionality

    def test_squared_dimensionality(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2")
        assert u.dimensionality == "[length] ** 2"

    def test_force_dimensionality(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("kilogram * meter / second ** 2")
        dim = u.dimensionality
        assert "[mass]" in dim
        assert "[length]" in dim
        assert "[time]" in dim

    def test_dimensionless_unit(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("dimensionless")
        assert u.dimensionless is True

    def test_not_dimensionless(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        assert u.dimensionless is False

    def test_power_zero_is_dimensionless(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter") ** 0
        assert u.dimensionless is True

    def test_prefixed_unit_same_dimensionality(self, ureg: UnitRegistry) -> None:
        m = ureg.parse_units("meter")
        km = ureg.parse_units("kilometer")
        assert m.dimensionality == km.dimensionality


# ---------------------------------------------------------------------------
# 5. TestUnitRepresentation
# ---------------------------------------------------------------------------


class TestUnitRepresentation:
    """String, repr, and format specifier output."""

    def test_str(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        assert str(u) == "m"

    def test_repr(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        r = repr(u)
        assert "Unit" in r
        assert "m" in r

    def test_units_str(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        assert u._units_str() == "m"

    def test_units_dict(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        d = u._units_dict()
        assert ("meter", 1.0) in d

    def test_units_dict_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        d = dict(u._units_dict())
        assert d["meter"] == pytest.approx(1.0)
        assert d["second"] == pytest.approx(-2.0)

    def test_format_pretty(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2 / second ** 2")
        fmt = format(u, "P")
        assert "\u00b2" in fmt  # superscript 2

    def test_format_pretty_abbreviated(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2 / second ** 2")
        fmt = format(u, "~P")
        assert "m" in fmt
        assert "s" in fmt
        assert "\u00b2" in fmt

    def test_format_latex(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2 / second ** 2")
        fmt = format(u, "L")
        assert r"\frac" in fmt
        assert "$" in fmt

    def test_format_latex_abbreviated(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2 / second ** 2")
        fmt = format(u, "~L")
        assert r"\frac" in fmt

    def test_format_html(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter ** 2")
        fmt = format(u, "H")
        assert "<sup>" in fmt

    def test_format_compact(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        fmt = format(u, "C")
        assert "meter" in fmt

    def test_format_default(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        fmt = format(u, "D")
        assert "meter" in fmt

    def test_format_dimensionless(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("dimensionless")
        s = str(u)
        assert "dimensionless" in s.lower() or s == ""

    def test_str_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        s = str(u)
        assert "/" in s
        assert "2" in s


# ---------------------------------------------------------------------------
# 6. TestUnitSerialization
# ---------------------------------------------------------------------------


class TestUnitSerialization:
    """Pickle, copy, and deepcopy round-trips."""

    def test_pickle_simple(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        data = pickle.dumps(u)
        u2 = pickle.loads(data)
        assert u == u2

    def test_pickle_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("kilogram * meter / second ** 2")
        data = pickle.dumps(u)
        u2 = pickle.loads(data)
        assert u == u2

    def test_pickle_preserves_hash(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        data = pickle.dumps(u)
        u2 = pickle.loads(data)
        assert hash(u) == hash(u2)

    def test_copy(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        u2 = copy.copy(u)
        assert u == u2

    def test_deepcopy(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        u2 = copy.deepcopy(u)
        assert u == u2

    def test_copy_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        u2 = copy.copy(u)
        assert u == u2

    def test_deepcopy_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("kilogram * meter ** 2 / second ** 2")
        u2 = copy.deepcopy(u)
        assert u == u2

    def test_pickle_dimensionless(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("dimensionless")
        data = pickle.dumps(u)
        u2 = pickle.loads(data)
        assert u2.dimensionless


# ---------------------------------------------------------------------------
# 7. TestUnitConversion
# ---------------------------------------------------------------------------


class TestUnitConversion:
    """Unit.from_() and Unit.m_from() conversions."""

    def test_from_simple(self, ureg: UnitRegistry) -> None:
        meter = ureg.Unit("meter")
        q = ureg.Quantity(1, "kilometer")
        result = meter.from_(q)
        assert result.magnitude == pytest.approx(1000.0)

    def test_from_preserves_target_unit(self, ureg: UnitRegistry) -> None:
        meter = ureg.Unit("meter")
        q = ureg.Quantity(5, "kilometer")
        result = meter.from_(q)
        assert str(result.units) == "m"

    def test_m_from_simple(self, ureg: UnitRegistry) -> None:
        meter = ureg.Unit("meter")
        q = ureg.Quantity(1, "kilometer")
        assert meter.m_from(q) == pytest.approx(1000.0)

    def test_from_incompatible_raises(self, ureg: UnitRegistry) -> None:
        meter = ureg.Unit("meter")
        q = ureg.Quantity(1, "second")
        with pytest.raises(DimensionalityError):
            meter.from_(q)

    def test_m_from_incompatible_raises(self, ureg: UnitRegistry) -> None:
        meter = ureg.Unit("meter")
        q = ureg.Quantity(1, "second")
        with pytest.raises(DimensionalityError):
            meter.m_from(q)

    def test_from_same_unit(self, ureg: UnitRegistry) -> None:
        meter = ureg.Unit("meter")
        q = ureg.Quantity(5.0, "meter")
        result = meter.from_(q)
        assert result.magnitude == pytest.approx(5.0)

    def test_m_from_prefixed(self, ureg: UnitRegistry) -> None:
        mm = ureg.Unit("millimeter")
        q = ureg.Quantity(1, "meter")
        assert mm.m_from(q) == pytest.approx(1000.0)

    def test_from_compound_unit(self, ureg: UnitRegistry) -> None:
        ms = ureg.Unit("meter / second")
        q = ureg.Quantity(1, "kilometer / hour")
        result = ms.from_(q)
        assert result.magnitude == pytest.approx(1.0 / 3.6, rel=1e-6)

    def test_systems_property(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        assert isinstance(u.systems, frozenset)
        assert "mks" in u.systems


# ---------------------------------------------------------------------------
# 8. TestRegistryUnitOps
# ---------------------------------------------------------------------------


class TestRegistryUnitOps:
    """Registry-level operations: define, get_name, get_symbol, etc."""

    def test_define_custom_unit(self, ureg: UnitRegistry) -> None:
        ureg.define("test_port_unit_a = 2 * meter")
        q = ureg.Quantity(1, "test_port_unit_a")
        result = q.to("meter")
        assert result.magnitude == pytest.approx(2.0)

    def test_define_with_alias(self, ureg: UnitRegistry) -> None:
        ureg.define("test_port_unit_b = 3 * meter = tpub")
        q = ureg.Quantity(1, "tpub")
        result = q.to("meter")
        assert result.magnitude == pytest.approx(3.0)

    def test_define_does_not_clobber_builtin(self, ureg: UnitRegistry) -> None:
        """Defining a unit that already exists should not break the registry."""
        original = ureg.Quantity(1, "meter").to("foot").magnitude
        ureg.define("test_port_shadow = 1 * meter")
        after = ureg.Quantity(1, "meter").to("foot").magnitude
        assert original == pytest.approx(after)

    def test_get_name(self, ureg: UnitRegistry) -> None:
        assert ureg.get_name("m") == "meter"

    def test_get_name_from_alias(self, ureg: UnitRegistry) -> None:
        name = ureg.get_name("metre")
        assert name == "meter"

    def test_get_symbol(self, ureg: UnitRegistry) -> None:
        sym = ureg.get_symbol("meter")
        assert sym == "m"

    def test_get_symbol_gram(self, ureg: UnitRegistry) -> None:
        sym = ureg.get_symbol("gram")
        assert sym == "g"

    def test_parse_unit_name_plain(self, ureg: UnitRegistry) -> None:
        result = ureg.parse_unit_name("meter")
        assert len(result) > 0
        assert result[0][1] == "meter"

    def test_parse_unit_name_prefixed(self, ureg: UnitRegistry) -> None:
        result = ureg.parse_unit_name("kilometer")
        assert len(result) > 0
        assert result[0][0] == "kilo"
        assert result[0][1] == "meter"

    def test_get_base_units(self, ureg: UnitRegistry) -> None:
        factor, base = ureg.get_base_units("kilometer")
        assert factor == pytest.approx(1000.0)
        assert isinstance(base, Unit)

    def test_get_base_units_compound(self, ureg: UnitRegistry) -> None:
        factor, _base = ureg.get_base_units("kilometer / hour")
        assert factor == pytest.approx(1000.0 / 3600.0, rel=1e-6)

    def test_get_compatible_units(self, ureg: UnitRegistry) -> None:
        compat = ureg.get_compatible_units("meter")
        assert isinstance(compat, list)
        assert "foot" in compat
        assert "inch" in compat

    def test_get_compatible_units_excludes_incompatible(
        self, ureg: UnitRegistry
    ) -> None:
        compat = ureg.get_compatible_units("meter")
        assert "second" not in compat
        assert "kilogram" not in compat

    def test_contains_known_unit(self, ureg: UnitRegistry) -> None:
        assert "meter" in ureg

    def test_contains_symbol(self, ureg: UnitRegistry) -> None:
        assert "m" in ureg

    def test_not_contains_unknown(self, ureg: UnitRegistry) -> None:
        assert "nonexistent_xyzzy_unit" not in ureg

    def test_iter_yields_unit_names(self, ureg: UnitRegistry) -> None:
        units = list(ureg)
        assert len(units) > 0
        assert "meter" in units
        assert "second" in units

    def test_load_definitions(self, ureg: UnitRegistry) -> None:
        ureg.load_definitions("test_port_loaded_unit = 42 * meter")
        assert "test_port_loaded_unit" in ureg
        q = ureg.Quantity(1, "test_port_loaded_unit")
        assert q.to("meter").magnitude == pytest.approx(42.0)

    def test_get_dimensionality(self, ureg: UnitRegistry) -> None:
        assert ureg.get_dimensionality("meter") == "[length]"

    def test_get_dimensionality_compound(self, ureg: UnitRegistry) -> None:
        dim = ureg.get_dimensionality("meter / second")
        assert "[length]" in dim
        assert "[time]" in dim

    def test_is_compatible_with_registry(self, ureg: UnitRegistry) -> None:
        assert ureg.is_compatible_with("meter", "foot")

    def test_is_not_compatible_with_registry(self, ureg: UnitRegistry) -> None:
        assert not ureg.is_compatible_with("meter", "second")

    def test_convert(self, ureg: UnitRegistry) -> None:
        result = ureg.convert(1.0, "kilometer", "meter")
        assert result == pytest.approx(1000.0)

    def test_parse_expression(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("5 meter")
        assert q.magnitude == pytest.approx(5.0)


# ---------------------------------------------------------------------------
# 9. TestUnitParsing
# ---------------------------------------------------------------------------


class TestUnitParsing:
    """Parsing of simple, prefixed, compound, plural, and alias forms."""

    @pytest.mark.parametrize(
        "unit_str",
        ["meter", "m", "metre", "meters", "metres"],
    )
    def test_meter_aliases(self, ureg: UnitRegistry, unit_str: str) -> None:
        u = ureg.parse_units(unit_str)
        assert u == ureg.parse_units("meter")

    @pytest.mark.parametrize(
        ("prefix", "base"),
        [
            ("kilo", "meter"),
            ("milli", "meter"),
            ("centi", "meter"),
            ("micro", "meter"),
            ("nano", "meter"),
        ],
    )
    def test_prefixed_units(
        self,
        ureg: UnitRegistry,
        prefix: str,
        base: str,
    ) -> None:
        u = ureg.parse_units(f"{prefix}{base}")
        assert u.dimensionality == ureg.parse_units(base).dimensionality

    @pytest.mark.parametrize(
        "expr",
        [
            "meter / second",
            "meter * second",
            "kilogram * meter / second ** 2",
            "meter ** 2",
            "meter ** 3",
        ],
    )
    def test_compound_expressions(self, ureg: UnitRegistry, expr: str) -> None:
        u = ureg.parse_units(expr)
        assert len(u._units_dict()) > 0

    def test_parse_newton(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("newton")
        dim = u.dimensionality
        assert "[mass]" in dim
        assert "[length]" in dim
        assert "[time]" in dim

    def test_parse_joule(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("joule")
        dim = u.dimensionality
        assert "[mass]" in dim
        assert "[length]" in dim

    def test_parse_watt(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("watt")
        dim = u.dimensionality
        assert "[mass]" in dim
        assert "[time]" in dim

    def test_parse_temperature_celsius(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("degree_Celsius")
        assert "[temperature]" in u.dimensionality

    def test_parse_temperature_fahrenheit(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("degree_Fahrenheit")
        assert "[temperature]" in u.dimensionality

    def test_parse_round_trip_simple(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        s = str(u)
        u2 = ureg.parse_units(s)
        assert u == u2

    def test_parse_round_trip_compound(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        s = str(u)
        u2 = ureg.parse_units(s)
        assert u == u2

    def test_parse_unit_with_per(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        d = dict(u._units_dict())
        assert d["second"] == pytest.approx(-1.0)

    @pytest.mark.parametrize(
        "unit_str",
        [
            "foot",
            "inch",
            "mile",
            "yard",
            "pound",
            "ounce",
        ],
    )
    def test_imperial_units_exist(self, ureg: UnitRegistry, unit_str: str) -> None:
        u = ureg.parse_units(unit_str)
        assert len(u._units_dict()) > 0



"""Port of pint's test_umath.py for pintrs.

Tests numpy ufunc integration with ArrayQuantity, covering math operations,
trigonometric functions, comparison functions, and floating-point functions.
"""

from __future__ import annotations

import pytest

np = pytest.importorskip("numpy")

from pintrs import UnitRegistry
from pintrs.numpy_support import ArrayQuantity

PI = np.pi


@pytest.fixture
def ureg() -> UnitRegistry:
    """Shared unit registry for all tests."""
    return UnitRegistry()


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _aq(
    values: list[float] | list[complex],
    units: str,
    ureg: UnitRegistry,
) -> ArrayQuantity:
    """Shorthand for creating an ArrayQuantity."""
    return ArrayQuantity(np.asarray(values), units, ureg)


# ---------------------------------------------------------------------------
# Math ufuncs
# ---------------------------------------------------------------------------


class TestMathUfuncs:
    """numpy math ufuncs applied to ArrayQuantity."""

    def test_add(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        result = np.add(a, b)
        np.testing.assert_allclose(result.magnitude, [3.0, 6.0, 9.0, 12.0])

    def test_subtract(self, ureg: UnitRegistry) -> None:
        a = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        b = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.subtract(a, b)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, 3.0, 4.0])

    def test_multiply(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        result = np.multiply(a, b)
        np.testing.assert_allclose(result.magnitude, [2.0, 8.0, 18.0, 32.0])

    def test_multiply_scalar(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.multiply(a, 3.0)
        np.testing.assert_allclose(result.magnitude, [3.0, 6.0, 9.0, 12.0])

    def test_divide(self, ureg: UnitRegistry) -> None:
        a = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        b = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.divide(a, b)
        np.testing.assert_allclose(result.magnitude, [2.0, 2.0, 2.0, 2.0])

    def test_true_divide(self, ureg: UnitRegistry) -> None:
        a = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        b = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.true_divide(a, b)
        np.testing.assert_allclose(result.magnitude, [2.0, 2.0, 2.0, 2.0])

    def test_floor_divide(self, ureg: UnitRegistry) -> None:
        a = _aq([5.0, 7.0, 10.0, 13.0], "joule", ureg)
        b = _aq([2.0, 3.0, 3.0, 4.0], "joule", ureg)
        result = np.floor_divide(a, b)
        np.testing.assert_allclose(result.magnitude, [2.0, 2.0, 3.0, 3.0])

    def test_negative(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.negative(q)
        np.testing.assert_allclose(result.magnitude, [-1.0, -2.0, -3.0, -4.0])

    def test_positive(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, -2.0, 3.0, -4.0], "joule", ureg)
        result = np.positive(q)
        np.testing.assert_allclose(result.magnitude, [1.0, -2.0, 3.0, -4.0])

    def test_absolute(self, ureg: UnitRegistry) -> None:
        q = _aq([-1.0, 2.0, -3.0, 4.0], "joule", ureg)
        result = np.absolute(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, 3.0, 4.0])

    def test_absolute_complex(self, ureg: UnitRegistry) -> None:
        q = _aq([1 + 1j, 2 + 2j, 3 + 3j, 4 + 4j], "meter", ureg)
        result = np.absolute(q)
        expected = np.absolute(np.array([1 + 1j, 2 + 2j, 3 + 3j, 4 + 4j]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_rint(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, 3.5, 4.1], "joule", ureg)
        result = np.rint(q)
        expected = np.rint(np.array([1.4, 2.6, 3.5, 4.1]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_conj(self, ureg: UnitRegistry) -> None:
        q = _aq([1 + 1j, 2 + 2j, 3 + 3j, 4 + 4j], "meter", ureg)
        result = np.conj(q)
        expected = np.conj(np.array([1 + 1j, 2 + 2j, 3 + 3j, 4 + 4j]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_exp(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "dimensionless", ureg)
        result = np.exp(q)
        expected = np.exp(np.array([1.0, 2.0, 3.0, 4.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_exp2(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "dimensionless", ureg)
        result = np.exp2(q)
        expected = np.exp2(np.array([1.0, 2.0, 3.0, 4.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_log(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "dimensionless", ureg)
        result = np.log(q)
        expected = np.log(np.array([1.0, 2.0, 3.0, 4.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_log2(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 4.0, 8.0], "dimensionless", ureg)
        result = np.log2(q)
        expected = np.log2(np.array([1.0, 2.0, 4.0, 8.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_log10(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 10.0, 100.0, 1000.0], "dimensionless", ureg)
        result = np.log10(q)
        expected = np.log10(np.array([1.0, 10.0, 100.0, 1000.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_expm1(self, ureg: UnitRegistry) -> None:
        q = _aq([0.0, 0.1, 0.5, 1.0], "dimensionless", ureg)
        result = np.expm1(q)
        expected = np.expm1(np.array([0.0, 0.1, 0.5, 1.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_sqrt(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 4.0, 9.0, 16.0], "m ** 2", ureg)
        result = np.sqrt(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, 3.0, 4.0])

    def test_sqrt_dimensionless(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 4.0, 9.0, 16.0], "dimensionless", ureg)
        result = np.sqrt(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, 3.0, 4.0])

    def test_square(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        result = np.square(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 4.0, 9.0, 16.0])

    def test_reciprocal(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 4.0, 5.0], "meter", ureg)
        result = np.reciprocal(q)
        expected = np.reciprocal(np.array([1.0, 2.0, 4.0, 5.0]))
        np.testing.assert_allclose(result.magnitude, expected)

    def test_remainder(self, ureg: UnitRegistry) -> None:
        a = _aq([5.0, 7.0, 10.0, 13.0], "joule", ureg)
        b = _aq([2.0, 3.0, 3.0, 4.0], "joule", ureg)
        result = np.remainder(a, b)
        expected = np.remainder(
            np.array([5.0, 7.0, 10.0, 13.0]),
            np.array([2.0, 3.0, 3.0, 4.0]),
        )
        np.testing.assert_allclose(result.magnitude, expected)

    def test_mod(self, ureg: UnitRegistry) -> None:
        a = _aq([5.0, 7.0, 10.0, 13.0], "joule", ureg)
        b = _aq([2.0, 3.0, 3.0, 4.0], "joule", ureg)
        result = np.mod(a, b)
        expected = np.mod(
            np.array([5.0, 7.0, 10.0, 13.0]),
            np.array([2.0, 3.0, 3.0, 4.0]),
        )
        np.testing.assert_allclose(result.magnitude, expected)

    def test_fmod(self, ureg: UnitRegistry) -> None:
        a = _aq([5.0, 7.0, 10.0, 13.0], "joule", ureg)
        b = _aq([2.0, 3.0, 3.0, 4.0], "joule", ureg)
        result = np.fmod(a, b)
        expected = np.fmod(
            np.array([5.0, 7.0, 10.0, 13.0]),
            np.array([2.0, 3.0, 3.0, 4.0]),
        )
        np.testing.assert_allclose(result.magnitude, expected)


# ---------------------------------------------------------------------------
# Trigonometric ufuncs
# ---------------------------------------------------------------------------


class TestTrigUfuncs:
    """numpy trigonometric ufuncs applied to ArrayQuantity."""

    def test_sin_dimensionless(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.sin(q)
        np.testing.assert_allclose(result.magnitude, np.sin(values))

    def test_sin_radian(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "radian", ureg)
        result = np.sin(q)
        np.testing.assert_allclose(result.magnitude, np.sin(values))

    def test_cos_dimensionless(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.cos(q)
        np.testing.assert_allclose(result.magnitude, np.cos(values))

    def test_cos_radian(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "radian", ureg)
        result = np.cos(q)
        np.testing.assert_allclose(result.magnitude, np.cos(values))

    def test_tan_dimensionless(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.tan(q)
        np.testing.assert_allclose(result.magnitude, np.tan(values))

    def test_tan_radian(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "radian", ureg)
        result = np.tan(q)
        np.testing.assert_allclose(result.magnitude, np.tan(values))

    def test_arcsin(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, 0.9, 0.1)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.arcsin(q)
        np.testing.assert_allclose(result.magnitude, np.arcsin(values))

    def test_arccos(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, 0.9, 0.1)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.arccos(q)
        np.testing.assert_allclose(result.magnitude, np.arccos(values))

    def test_arctan(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, 0.9, 0.1)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.arctan(q)
        np.testing.assert_allclose(result.magnitude, np.arctan(values))

    def test_arctan2(self, ureg: UnitRegistry) -> None:
        y_vals = np.arange(0, 0.9, 0.1)
        x_vals = np.arange(0.9, 0.0, -0.1)
        y = _aq(y_vals.tolist(), "meter", ureg)
        x = _aq(x_vals.tolist(), "meter", ureg)
        result = np.arctan2(y, x)
        np.testing.assert_allclose(
            result.magnitude,
            np.arctan2(y_vals, x_vals),
        )

    def test_hypot(self, ureg: UnitRegistry) -> None:
        a = _aq([3.0], "meter", ureg)
        b = _aq([4.0], "meter", ureg)
        result = np.hypot(a, b)
        np.testing.assert_allclose(result.magnitude, [5.0])

    def test_sinh(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.sinh(q)
        np.testing.assert_allclose(result.magnitude, np.sinh(values))

    def test_sinh_radian(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "radian", ureg)
        result = np.sinh(q)
        np.testing.assert_allclose(result.magnitude, np.sinh(values))

    def test_cosh(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.cosh(q)
        np.testing.assert_allclose(result.magnitude, np.cosh(values))

    def test_cosh_radian(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "radian", ureg)
        result = np.cosh(q)
        np.testing.assert_allclose(result.magnitude, np.cosh(values))

    def test_tanh(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.tanh(q)
        np.testing.assert_allclose(result.magnitude, np.tanh(values))

    def test_tanh_radian(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, PI / 2, PI / 4)
        q = _aq(values.tolist(), "radian", ureg)
        result = np.tanh(q)
        np.testing.assert_allclose(result.magnitude, np.tanh(values))

    def test_arcsinh(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, 0.9, 0.1)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.arcsinh(q)
        np.testing.assert_allclose(result.magnitude, np.arcsinh(values))

    def test_arccosh(self, ureg: UnitRegistry) -> None:
        values = np.arange(1.0, 1.9, 0.1)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.arccosh(q)
        np.testing.assert_allclose(result.magnitude, np.arccosh(values))

    def test_arctanh(self, ureg: UnitRegistry) -> None:
        values = np.arange(0, 0.9, 0.1)
        q = _aq(values.tolist(), "dimensionless", ureg)
        result = np.arctanh(q)
        np.testing.assert_allclose(result.magnitude, np.arctanh(values))


# ---------------------------------------------------------------------------
# Comparison ufuncs
# ---------------------------------------------------------------------------


class TestComparisonUfuncs:
    """numpy comparison ufuncs applied to ArrayQuantity."""

    def test_greater(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        result = np.greater(a, b)
        expected = np.greater(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([2.0, 4.0, 6.0, 8.0]),
        )
        np.testing.assert_array_equal(result, expected)

    def test_greater_equal(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([1.0, 4.0, 3.0, 8.0], "joule", ureg)
        result = np.greater_equal(a, b)
        expected = np.greater_equal(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([1.0, 4.0, 3.0, 8.0]),
        )
        np.testing.assert_array_equal(result, expected)

    def test_less(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        result = np.less(a, b)
        expected = np.less(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([2.0, 4.0, 6.0, 8.0]),
        )
        np.testing.assert_array_equal(result, expected)

    def test_less_equal(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([1.0, 4.0, 3.0, 8.0], "joule", ureg)
        result = np.less_equal(a, b)
        expected = np.less_equal(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([1.0, 4.0, 3.0, 8.0]),
        )
        np.testing.assert_array_equal(result, expected)

    def test_not_equal(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([1.0, 4.0, 3.0, 8.0], "joule", ureg)
        result = np.not_equal(a, b)
        expected = np.not_equal(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([1.0, 4.0, 3.0, 8.0]),
        )
        np.testing.assert_array_equal(result, expected)

    def test_equal(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([1.0, 4.0, 3.0, 8.0], "joule", ureg)
        result = np.equal(a, b)
        expected = np.equal(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([1.0, 4.0, 3.0, 8.0]),
        )
        np.testing.assert_array_equal(result, expected)


# ---------------------------------------------------------------------------
# Floating ufuncs
# ---------------------------------------------------------------------------


class TestFloatingUfuncs:
    """numpy floating-point ufuncs applied to ArrayQuantity."""

    def test_isfinite(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.isfinite(q)
        np.testing.assert_array_equal(result, [True, True, True, True])

    def test_isinf(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.isinf(q)
        np.testing.assert_array_equal(result, [False, False, False, False])

    def test_isnan(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        result = np.isnan(q)
        np.testing.assert_array_equal(result, [False, False, False, False])

    def test_signbit(self, ureg: UnitRegistry) -> None:
        q = _aq([1.0, -2.0, 3.0, -4.0], "joule", ureg)
        result = np.signbit(q)
        np.testing.assert_array_equal(result, [False, True, False, True])

    def test_copysign(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([-1.0, 1.0, -1.0, 1.0], "joule", ureg)
        result = np.copysign(a, b)
        np.testing.assert_allclose(result.magnitude, [-1.0, 2.0, -3.0, 4.0])

    def test_nextafter(self, ureg: UnitRegistry) -> None:
        a = _aq([1.0, 2.0, 3.0, 4.0], "joule", ureg)
        b = _aq([2.0, 4.0, 6.0, 8.0], "joule", ureg)
        result = np.nextafter(a, b)
        expected = np.nextafter(
            np.array([1.0, 2.0, 3.0, 4.0]),
            np.array([2.0, 4.0, 6.0, 8.0]),
        )
        np.testing.assert_allclose(result.magnitude, expected)

    def test_floor(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, 3.1, 4.9], "joule", ureg)
        result = np.floor(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, 3.0, 4.0])

    def test_ceil(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, 3.1, 4.9], "joule", ureg)
        result = np.ceil(q)
        np.testing.assert_allclose(result.magnitude, [2.0, 3.0, 4.0, 5.0])

    def test_trunc(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, -3.1, -4.9], "joule", ureg)
        result = np.trunc(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, -3.0, -4.0])

    def test_floor_meter(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, 3.1, 4.9], "meter", ureg)
        result = np.floor(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, 3.0, 4.0])

    def test_ceil_meter(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, 3.1, 4.9], "meter", ureg)
        result = np.ceil(q)
        np.testing.assert_allclose(result.magnitude, [2.0, 3.0, 4.0, 5.0])

    def test_trunc_dimensionless(self, ureg: UnitRegistry) -> None:
        q = _aq([1.4, 2.6, -3.1, -4.9], "dimensionless", ureg)
        result = np.trunc(q)
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0, -3.0, -4.0])



"""Formatter tests - ported from pint's test_formatter.py.

Tests formatting behavior through quantity/unit __format__ and str()
rather than through pint's internal formatter functions.
"""

from __future__ import annotations

import pytest
from pintrs import UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestQuantityDefaultFormat:
    """Test default formatting of quantities."""

    def test_str_simple(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        s = str(q)
        assert "10" in s
        assert "m" in s

    def test_format_empty_spec(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.5, "meter")
        assert format(q, "") == str(q)

    def test_dimensionless_str(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(42, "dimensionless")
        s = str(q)
        assert "42" in s

    def test_compound_unit_str(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(9.81, "meter / second ** 2")
        s = str(q)
        assert "9.81" in s


class TestQuantityMagnitudeFormat:
    """Test magnitude formatting specifiers."""

    def test_fixed_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14159, "meter")
        result = format(q, ".2f")
        assert "3.14" in result

    def test_scientific_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(12345, "meter")
        result = format(q, ".2e")
        assert "1.23" in result or "1.24" in result


class TestUnitFormatStrings:
    """Test unit format string specifiers (~, P, L, H, D, C)."""

    def test_abbreviated_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "~")
        assert "m" in result

    def test_pretty_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter / second")
        result = format(q, "~P")
        assert "m" in result
        assert "s" in result

    def test_pretty_exponents(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "meter / second ** 2")
        result = format(q, "~P")
        assert "2" in result or "\u00b2" in result

    def test_latex_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "~L")
        assert "$" in result or "mathrm" in result

    def test_latex_compound(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter / second ** 2")
        result = format(q, "~L")
        assert "frac" in result or "^" in result

    def test_html_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "~H")
        assert "m" in result

    def test_html_exponents(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "meter / second ** 2")
        result = format(q, "~H")
        assert "sup" in result or "2" in result

    def test_default_full_names(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "D")
        assert "meter" in result

    def test_compact_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5000, "meter")
        result = format(q, "~C")
        assert "5" in result
        assert "km" in result or "kilometer" in result


class TestUnitFormat:
    """Test formatting of Unit objects directly."""

    def test_unit_str(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        result = str(u)
        assert result != ""

    def test_unit_format_default(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        assert format(u, "") == str(u)

    def test_unit_format_pretty(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        result = format(u, "~P")
        assert "m" in result

    def test_unit_format_latex(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        result = format(u, "~L")
        assert "$" in result or "mathrm" in result

    def test_unit_format_html(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        result = format(u, "~H")
        assert "m" in result

    def test_unit_format_abbreviated(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        result = format(u, "~")
        assert "m" in result


class TestFormatEdgeCases:
    """Test formatting edge cases."""

    def test_zero_quantity(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "meter")
        result = format(q, "~P")
        assert "0" in result

    def test_negative_quantity(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(-5, "meter")
        result = format(q, ".1f~P")
        assert "-5.0" in result

    def test_power_unit(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "meter ** 2")
        result = format(q, "~P")
        assert "100" in result

    def test_combined_magnitude_and_unit_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        result = format(q, ".2f~P")
        assert "3.14" in result

    def test_latex_with_magnitude(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        result = format(q, ".2f~L")
        assert "3.14" in result



"""Tests for logarithmic units."""

from __future__ import annotations

import pytest
from pintrs import LogarithmicQuantity, UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestLogarithmicCreation:
    def test_create_dB(self) -> None:
        q = LogarithmicQuantity(3.0, "dB")
        assert q.magnitude == 3.0
        assert q.units == "dB"

    def test_create_dBm(self) -> None:
        q = LogarithmicQuantity(30.0, "dBm")
        assert q.magnitude == 30.0
        assert q.units == "dBm"

    def test_create_Np(self) -> None:
        q = LogarithmicQuantity(1.0, "Np")
        assert q.units == "Np"

    def test_invalid_unit(self) -> None:
        with pytest.raises(ValueError, match="Unknown logarithmic unit"):
            LogarithmicQuantity(1.0, "dBx")


class TestLinearConversion:
    def test_dB_to_linear(self) -> None:
        q = LogarithmicQuantity(10.0, "dB")
        assert abs(q.to_linear() - 10.0) < 1e-10

    def test_dB_to_linear_20(self) -> None:
        q = LogarithmicQuantity(20.0, "dB")
        assert abs(q.to_linear() - 100.0) < 1e-10

    def test_dB_zero(self) -> None:
        q = LogarithmicQuantity(0.0, "dB")
        assert abs(q.to_linear() - 1.0) < 1e-10

    def test_dBm_to_linear(self) -> None:
        q = LogarithmicQuantity(30.0, "dBm")
        # 30 dBm = 1 W = 1000 mW
        assert abs(q.to_linear() - 1000.0) < 1e-6

    def test_from_linear(self) -> None:
        q = LogarithmicQuantity.from_linear(100.0, "dB")
        assert abs(q.magnitude - 20.0) < 1e-10

    def test_from_linear_negative_raises(self) -> None:
        with pytest.raises(ValueError, match="non-positive"):
            LogarithmicQuantity.from_linear(-1.0, "dB")


class TestLogarithmicConversion:
    def test_dB_to_Bel(self) -> None:
        q = LogarithmicQuantity(20.0, "dB")
        result = q.to("Bel")
        assert abs(result.magnitude - 2.0) < 1e-10

    def test_dBm_to_dBW(self) -> None:
        q = LogarithmicQuantity(30.0, "dBm")
        result = q.to("dBW")
        assert abs(result.magnitude - 0.0) < 1e-10


class TestToQuantity:
    def test_dBm_to_quantity(self, ureg: UnitRegistry) -> None:
        q = LogarithmicQuantity(30.0, "dBm", ureg)
        result = q.to_quantity()
        # 30 dBm = 1 W
        assert abs(result.magnitude - 1.0) < 1e-6
        assert str(result.units) == "W"

    def test_dBW_to_quantity(self, ureg: UnitRegistry) -> None:
        q = LogarithmicQuantity(0.0, "dBW", ureg)
        result = q.to_quantity()
        assert abs(result.magnitude - 1.0) < 1e-6

    def test_plain_dB_raises(self) -> None:
        q = LogarithmicQuantity(10.0, "dB")
        with pytest.raises(ValueError, match="relative unit"):
            q.to_quantity()


class TestLogarithmicArithmetic:
    def test_add(self) -> None:
        a = LogarithmicQuantity(10.0, "dB")
        b = LogarithmicQuantity(3.0, "dB")
        result = a + b
        assert abs(result.magnitude - 13.0) < 1e-10

    def test_sub(self) -> None:
        a = LogarithmicQuantity(20.0, "dBm")
        b = LogarithmicQuantity(10.0, "dBm")
        result = a - b
        assert abs(result.magnitude - 10.0) < 1e-10

    def test_add_scalar(self) -> None:
        q = LogarithmicQuantity(10.0, "dB")
        result = q + 3.0
        assert abs(result.magnitude - 13.0) < 1e-10

    def test_neg(self) -> None:
        q = LogarithmicQuantity(10.0, "dB")
        assert (-q).magnitude == -10.0

    def test_eq(self) -> None:
        a = LogarithmicQuantity(10.0, "dB")
        b = LogarithmicQuantity(10.0, "dB")
        assert a == b

    def test_lt(self) -> None:
        a = LogarithmicQuantity(3.0, "dB")
        b = LogarithmicQuantity(10.0, "dB")
        assert a < b

    def test_repr(self) -> None:
        q = LogarithmicQuantity(10.0, "dBm")
        assert "10.0" in repr(q)
        assert "dBm" in repr(q)

    def test_str(self) -> None:
        q = LogarithmicQuantity(10.0, "dBm")
        assert str(q) == "10.0 dBm"



"""Infer base unit and to_compact tests - ported from pint's test_infer_base_unit.py.

Tests to_compact() through the quantity API. infer_base_unit is not
exposed by pintrs, so those tests are adapted to verify equivalent
behavior through conversions and to_base_units().
"""

from __future__ import annotations

import pytest
from pintrs import UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestBaseUnitInference:
    """Test that compound prefix units reduce to correct base units."""

    def test_mm_times_nm_reduces_to_m2(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "millimeter * nanometer")
        base = q.to_base_units()
        expected = ureg.Quantity(1e-12, "meter ** 2")
        assert abs(base.magnitude - expected.magnitude) < 1e-22
        assert str(base.units) == str(expected.units)

    def test_units_cancelling_to_seconds(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "meter") * ureg.Quantity(1, "millimeter")
        q = q / ureg.Quantity(1, "meter") / ureg.Quantity(1, "micrometer")
        q = q * ureg.Quantity(1, "second")
        base = q.to_base_units()
        assert base.is_compatible_with("second")

    def test_volt_prefixes(self, ureg: UnitRegistry) -> None:
        r = ureg.Quantity(1, "V") * ureg.Quantity(1, "mV") / ureg.Quantity(1, "kV")
        base = r.to_base_units()
        expected = ureg.Quantity(1, "V").to_base_units()
        assert str(base.units) == str(expected.units)
        assert r.is_compatible_with("V")
        assert abs(r.to("uV").magnitude - 1.0) < 1e-6


class TestToCompact:
    """Test to_compact() method."""

    def test_compact_small(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.001, "meter")
        compact = q.to_compact()
        assert abs(compact.magnitude - 1.0) < 1e-10
        assert "milli" in str(compact) or "mm" in str(compact)

    def test_compact_large(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5000, "meter")
        compact = q.to_compact()
        assert abs(compact.magnitude - 5.0) < 1e-10
        assert "km" in str(compact) or "kilo" in str(compact)

    def test_compact_very_large(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1500, "meter")
        compact = q.to_compact()
        assert abs(compact.magnitude - 1.5) < 1e-10

    def test_compact_micro(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.000001, "meter")
        compact = q.to_compact()
        assert abs(compact.magnitude - 1.0) < 1e-4

    def test_compact_preserves_value(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1000000000, "meter")
        compact = q.to_compact()
        roundtrip = compact.to("meter")
        assert abs(roundtrip.magnitude - 1000000000) < 1e-3

    def test_compact_compound_units(self, ureg: UnitRegistry) -> None:
        r = (
            ureg.Quantity(1000000000, "m")
            * ureg.Quantity(1, "mm")
            / ureg.Quantity(1, "s")
            / ureg.Quantity(1, "ms")
        )
        compact = r.to_compact()
        expected_in_base = r.to_base_units()
        compact_in_base = compact.to_base_units()
        assert abs(compact_in_base.magnitude - expected_in_base.magnitude) < 1e-3

    def test_compact_already_good(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5, "meter")
        compact = q.to_compact()
        roundtrip = compact.to("meter")
        assert abs(roundtrip.magnitude - 5.0) < 1e-10

    def test_compact_dimensionless(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(42, "dimensionless")
        compact = q.to_compact()
        assert abs(compact.magnitude - 42) < 1e-10


class TestToCompactWithVolts:
    """Test to_compact with voltage units."""

    def test_millivolt_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.001, "V")
        compact = q.to_compact()
        assert abs(compact.magnitude - 1.0) < 1e-10

    def test_kilovolt_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1000, "V")
        compact = q.to_compact()
        assert abs(compact.magnitude - 1.0) < 1e-10



"""Formatting tests - ported from pint's test_formatting.py."""

from __future__ import annotations

import pytest
from pintrs import UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestQuantityFormatDefault:
    def test_no_spec(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        assert str(q) == format(q, "")

    def test_magnitude_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14159, "meter")
        result = format(q, ".2f")
        assert "3.14" in result

    def test_magnitude_scientific(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(12345, "meter")
        result = format(q, ".2e")
        assert "1.23" in result or "1.24" in result


class TestQuantityFormatPretty:
    def test_pretty_simple(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter / second")
        result = format(q, "~P")
        assert "m" in result
        assert "s" in result

    def test_pretty_with_magnitude_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        result = format(q, ".2f~P")
        assert "3.14" in result

    def test_pretty_exponents_use_superscript(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "meter / second ** 2")
        result = format(q, "~P")
        assert "²" in result or "2" in result


class TestQuantityFormatLatex:
    def test_latex_simple(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "~L")
        assert "$" in result or "mathrm" in result

    def test_latex_compound(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter / second ** 2")
        result = format(q, "~L")
        assert "frac" in result or "^" in result

    def test_latex_with_magnitude(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        result = format(q, ".2f~L")
        assert "3.14" in result


class TestQuantityFormatHTML:
    def test_html_simple(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "~H")
        assert "m" in result

    def test_html_exponents_use_sup(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "meter / second ** 2")
        result = format(q, "~H")
        assert "sup" in result or "2" in result


class TestQuantityFormatCompact:
    def test_compact_large(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5000, "meter")
        result = format(q, "~C")
        assert "5" in result
        assert "km" in result or "kilometer" in result

    def test_compact_small(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.001, "meter")
        result = format(q, "~C")
        assert "mm" in result or "millimeter" in result or "1" in result


class TestQuantityFormatAbbreviation:
    def test_tilde_uses_abbreviations(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        abbreviated = format(q, "~")
        full = format(q, "")
        assert len(abbreviated) <= len(full) or abbreviated == full

    def test_tilde_pretty(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "~P")
        assert "m" in result

    def test_no_tilde_full_names(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = format(q, "D")
        assert "meter" in result


class TestUnitFormat:
    def test_unit_format_default(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second")
        result = format(u, "")
        assert str(u) == result

    def test_unit_format_pretty(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        result = format(u, "~P")
        assert "m" in result

    def test_unit_format_latex(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        result = format(u, "~L")
        assert "$" in result or "mathrm" in result

    def test_unit_format_html(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        result = format(u, "~H")
        assert "m" in result

    def test_unit_format_abbreviation(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        abbreviated = format(u, "~")
        assert "m" in abbreviated


class TestFormatEdgeCases:
    def test_dimensionless_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(42, "dimensionless")
        result = format(q, "~P")
        assert "42" in result

    def test_compound_unit_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(9.81, "meter / second ** 2")
        result = format(q, ".2f~P")
        assert "9.81" in result

    def test_power_unit_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "meter ** 2")
        result = format(q, "~P")
        assert "100" in result

    def test_zero_quantity_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "meter")
        result = format(q, "~P")
        assert "0" in result

    def test_negative_quantity_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(-5, "meter")
        result = format(q, ".1f~P")
        assert "-5.0" in result



"""Utility tests - ported from pint's test_util.py.

Tests internal utility behavior through the public API, since pintrs
implements these utilities in Rust and doesn't expose UnitsContainer,
ParserHelper, etc. directly.
"""

from __future__ import annotations

import pytest
from pintrs import UndefinedUnitError, UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


# ---------------------------------------------------------------------------
# Unit container behavior (tested through Unit and Quantity)
# ---------------------------------------------------------------------------


class TestUnitContainerBehavior:
    """Mirrors pint's TestUnitsContainer via the public Unit/Quantity API."""

    def test_unit_creation(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter * second ** 2")
        assert str(u) is not None

    def test_dimensionless_str(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter") / ureg.Quantity(1.0, "meter")
        assert q.dimensionless

    def test_unit_equality(self, ureg: UnitRegistry) -> None:
        u1 = ureg.parse_units("meter")
        u2 = ureg.parse_units("meter")
        assert u1 == u2

    def test_unit_inequality(self, ureg: UnitRegistry) -> None:
        u1 = ureg.parse_units("meter")
        u2 = ureg.parse_units("second")
        assert u1 != u2

    def test_unit_multiplication(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(2, "meter")
        q2 = ureg.Quantity(3, "second")
        result = q1 * q2
        assert result.magnitude == pytest.approx(6.0)

    def test_unit_division(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(6, "meter")
        q2 = ureg.Quantity(2, "second")
        result = q1 / q2
        assert result.magnitude == pytest.approx(3.0)

    def test_unit_power(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3, "meter")
        result = q**2
        assert result.magnitude == pytest.approx(9.0)


# ---------------------------------------------------------------------------
# String preprocessing (tested via parse_expression)
# ---------------------------------------------------------------------------


class TestStringPreprocessing:
    """Mirrors pint's TestStringProcessor by testing parsing behavior."""

    def test_squared_cubed(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1 meter^3")
        assert q.magnitude == pytest.approx(1.0)
        q2 = ureg.parse_expression("1 meter**3")
        assert q.to_base_units().magnitude == pytest.approx(
            q2.to_base_units().magnitude
        )

    def test_per(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1 meter / second")
        assert q.magnitude == pytest.approx(1.0)

    def test_scientific_notation(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1e3 meter")
        assert q.magnitude == pytest.approx(1000.0)

        q = ureg.parse_expression("1E-3 meter")
        assert q.magnitude == pytest.approx(0.001)

        q = ureg.parse_expression("1.5e2 meter")
        assert q.magnitude == pytest.approx(150.0)

    def test_space_multiplication(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1 meter")
        assert q.magnitude == pytest.approx(1.0)

        q = ureg.parse_expression("2.5 meter")
        assert q.magnitude == pytest.approx(2.5)

    def test_compound_units(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("9.81 meter / second ** 2")
        assert q.magnitude == pytest.approx(9.81)

        q = ureg.parse_expression("1 kilogram * meter / second ** 2")
        assert q.magnitude == pytest.approx(1.0)


# ---------------------------------------------------------------------------
# Parse expression evaluation (mirrors TestParseHelper)
# ---------------------------------------------------------------------------


class TestParseExpression:
    """Test expression parsing through the public API."""

    def test_simple_value(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("42")
        assert q.magnitude == pytest.approx(42.0)

    def test_value_with_units(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("4.2 meter")
        assert q.magnitude == pytest.approx(4.2)

    def test_compound_expression(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1 meter * second")
        assert q.magnitude == pytest.approx(1.0)

    def test_negative_exponent(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1 meter ** -1")
        assert q.magnitude == pytest.approx(1.0)

    def test_division_expression(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("1 meter / second")
        assert q.magnitude == pytest.approx(1.0)

    def test_from_string(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity("4.2 meter")
        assert q.magnitude == pytest.approx(4.2)

    def test_nan_parsing(self, ureg: UnitRegistry) -> None:
        import math

        q = ureg.Quantity(float("nan"), "meter")
        assert math.isnan(q.magnitude)


# ---------------------------------------------------------------------------
# Quantity iteration and sizing
# ---------------------------------------------------------------------------


class TestIterableAndSized:
    """Mirrors pint's TestOtherUtils for iterable/sized behavior."""

    def test_quantity_not_iterable(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(42, "meter")
        with pytest.raises(TypeError):
            iter(q)  # type: ignore[call-overload]

    def test_quantity_has_no_len(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(42, "meter")
        with pytest.raises(TypeError):
            len(q)  # type: ignore[arg-type]


# ---------------------------------------------------------------------------
# Define custom units (mirrors TestDefinition behavior)
# ---------------------------------------------------------------------------


class TestDefine:
    """Test unit definition through the public API."""

    def test_define_simple_unit(self, ureg: UnitRegistry) -> None:
        ureg.define("testunit = 42 * meter")
        q = ureg.Quantity(1, "testunit")
        result = q.to("meter")
        assert result.magnitude == pytest.approx(42.0)

    def test_define_with_prefix(self, ureg: UnitRegistry) -> None:
        ureg.define("myfoot = 0.3048 * meter")
        q = ureg.Quantity(1, "myfoot")
        result = q.to("meter")
        assert result.magnitude == pytest.approx(0.3048)

    def test_undefined_unit_raises(self, ureg: UnitRegistry) -> None:
        with pytest.raises(UndefinedUnitError):
            ureg.Quantity(1, "nonexistent_unit_xyz")


# ---------------------------------------------------------------------------
# Unit formatting (mirrors TestFormatter)
# ---------------------------------------------------------------------------


class TestUnitFormatting:
    """Test unit/quantity formatting through the public API."""

    def test_quantity_str(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        s = str(q)
        assert "3.14" in s
        assert "m" in s

    def test_quantity_repr(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14, "meter")
        r = repr(q)
        assert "3.14" in r

    def test_unit_str(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        s = str(u)
        assert "m" in s
        assert "s" in s

    def test_format_pretty(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter / second ** 2")
        s = format(u, "P")
        assert "²" in s

    def test_format_latex(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter / second")
        s = format(q, "L")
        assert "$" in s or "frac" in s

    def test_dimensionless_format(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter") / ureg.Quantity(1.0, "meter")
        s = str(q)
        assert "1" in s


# ---------------------------------------------------------------------------
# to_compact (mirrors test_infer_base_unit.test_to_compact)
# ---------------------------------------------------------------------------


class TestToCompact:
    """Test to_compact behavior through the public API."""

    def test_large_value_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1e9, "meter")
        c = q.to_compact()
        assert c.magnitude == pytest.approx(1.0)
        assert "Gm" in str(c) or "gigameter" in str(c).lower() or "gm" in str(c)

    def test_small_value_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.001, "meter")
        c = q.to_compact()
        assert c.magnitude == pytest.approx(1.0)

    def test_already_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "meter")
        c = q.to_compact()
        # to_compact may choose a different prefix, but the value is equivalent
        assert c.to("meter").magnitude == pytest.approx(5.0)

    def test_compound_unit_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1000, "meter / second")
        c = q.to_compact()
        assert c.magnitude >= 0.1
        assert c.magnitude <= 1000


# ---------------------------------------------------------------------------
# Volts / combined unit inference
# ---------------------------------------------------------------------------


class TestCombinedUnits:
    """Test combined unit operations (mirrors test_infer_base_unit.test_volts)."""

    def test_volt_operations(self, ureg: UnitRegistry) -> None:
        r = ureg.Quantity(1, "V") * ureg.Quantity(1, "mV") / ureg.Quantity(1, "kV")
        expected = ureg.Quantity(1e-6, "V")
        assert r.to("V").magnitude == pytest.approx(expected.magnitude, rel=1e-6)

    def test_meter_combinations(self, ureg: UnitRegistry) -> None:
        r = ureg.Quantity(1, "mm") * ureg.Quantity(1, "nm")
        expected = ureg.Quantity(1e-12, "meter ** 2")
        assert r.to("meter ** 2").magnitude == pytest.approx(
            expected.magnitude, rel=1e-6
        )



"""Run a subset of pint's test patterns against pintrs.

This tests the most common pint usage patterns to verify drop-in compatibility.
We can't run pint's full test suite directly because it imports pint internals,
but we replicate the key test patterns here using pintrs.
"""

from __future__ import annotations

import copy
import datetime
import math
import pickle

import pytest
from pintrs import (
    DimensionalityError,
    Measurement,
    Quantity,
    Unit,
    UnitRegistry,
)


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


# === Pint test_quantity.py patterns ===


class TestQuantityCreation:
    """Mirrors pint's TestQuantity.test_quantity_creation."""

    def test_number_and_string(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(4.2, "meter")
        assert q.magnitude == 4.2

    def test_string(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity("4.2*meter")
        assert abs(q.magnitude - 4.2) < 1e-10

    def test_quantity_from_quantity(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(4.2, "meter")
        q2 = ureg.Quantity(q1)
        assert q2.magnitude == 4.2

    def test_quantity_from_quantity_convert(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(4.2, "meter")
        q2 = ureg.Quantity(q1, "centimeter")
        assert abs(q2.magnitude - 420.0) < 1e-10

    def test_nan(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(float("nan"), "meter")
        assert math.isnan(q.magnitude)

    def test_inf(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(float("inf"), "meter")
        assert math.isinf(q.magnitude)

    def test_empty_string_raises(self, ureg: UnitRegistry) -> None:
        with pytest.raises(ValueError):
            ureg.Quantity("")


class TestQuantityToFromTuple:
    """Mirrors pint's test_quantity_to_from_tuple."""

    def test_to_tuple(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(4.2, "meter")
        mag, items = q.to_tuple()
        assert mag == 4.2
        assert items[0][0] == "meter"
        assert items[0][1] == 1.0

    def test_from_tuple(self) -> None:
        q = Quantity.from_tuple((4.2, [("meter", 1.0)]))
        assert abs(q.magnitude - 4.2) < 1e-10


class TestQuantityConversion:
    """Mirrors pint's conversion tests."""

    def test_convert(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "inch")
        result = q.to("centimeter")
        assert abs(result.magnitude - 2.54) < 1e-6

    def test_convert_from(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(2.0, "inch")
        assert abs(q.m_as("meter") - 0.0508) < 1e-6

    def test_incompatible_raises(self, ureg: UnitRegistry) -> None:
        with pytest.raises(DimensionalityError):
            ureg.Quantity(1.0, "meter").to("second")

    def test_to_base_units(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "kilometer")
        base = q.to_base_units()
        assert abs(base.magnitude - 5000.0) < 1e-10
        assert str(base.units) == "m"

    def test_ito(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "kilometer")
        q.ito("meter")
        assert abs(q.magnitude - 1000.0) < 1e-10

    def test_to_compact(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0.001, "meter")
        compact = q.to_compact()
        assert abs(compact.magnitude - 1.0) < 1e-10
        assert "milli" in str(compact) or "mm" in str(compact)

    def test_to_unprefixed(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5, "km")
        result = q.to_unprefixed()
        assert abs(result.magnitude - 5000) < 1e-10


class TestQuantityArithmetic:
    """Mirrors pint's arithmetic tests."""

    def test_add_same_units(self, ureg: UnitRegistry) -> None:
        a = ureg.Quantity(1.0, "meter")
        b = ureg.Quantity(2.0, "meter")
        assert (a + b).magnitude == 3.0

    def test_add_compatible(self, ureg: UnitRegistry) -> None:
        a = ureg.Quantity(1.0, "kilometer")
        b = ureg.Quantity(25.0, "meter")
        result = a + b
        assert abs(result.magnitude - 1.025) < 1e-10

    def test_add_incompatible(self, ureg: UnitRegistry) -> None:
        with pytest.raises(DimensionalityError):
            ureg.Quantity(1.0, "meter") + ureg.Quantity(1.0, "second")

    def test_sub(self, ureg: UnitRegistry) -> None:
        a = ureg.Quantity(10.0, "meter")
        b = ureg.Quantity(3.0, "meter")
        assert (a - b).magnitude == 7.0

    def test_mul(self, ureg: UnitRegistry) -> None:
        a = ureg.Quantity(2.0, "meter")
        b = ureg.Quantity(3.0, "second")
        result = a * b
        assert result.magnitude == 6.0

    def test_mul_scalar(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "meter")
        assert (q * 3).magnitude == 15.0
        assert (3 * q).magnitude == 15.0
        assert (3.0 * q).magnitude == 15.0

    def test_div(self, ureg: UnitRegistry) -> None:
        a = ureg.Quantity(10.0, "meter")
        b = ureg.Quantity(2.0, "second")
        result = a / b
        assert result.magnitude == 5.0

    def test_div_scalar(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10.0, "meter")
        assert (q / 2).magnitude == 5.0

    def test_rdiv(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(2.0, "meter")
        result = 10.0 / q
        assert result.magnitude == 5.0

    def test_pow(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.0, "meter")
        result = q**2
        assert result.magnitude == 9.0

    def test_neg(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "meter")
        assert (-q).magnitude == -5.0

    def test_abs(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(-5.0, "meter")
        assert abs(q).magnitude == 5.0

    def test_round(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.14159, "meter")
        assert abs(round(q, 2).magnitude - 3.14) < 1e-10


class TestQuantityComparison:
    """Mirrors pint's comparison tests."""

    def test_eq(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") == ureg.Quantity(1.0, "meter")

    def test_eq_different_units(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "kilometer") == ureg.Quantity(1000.0, "meter")

    def test_ne(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") != ureg.Quantity(2.0, "meter")

    def test_lt(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") < ureg.Quantity(2.0, "meter")

    def test_gt(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(2.0, "meter") > ureg.Quantity(1.0, "meter")

    def test_le(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(1.0, "meter") <= ureg.Quantity(1.0, "meter")
        assert ureg.Quantity(1.0, "meter") <= ureg.Quantity(2.0, "meter")

    def test_ge(self, ureg: UnitRegistry) -> None:
        assert ureg.Quantity(2.0, "meter") >= ureg.Quantity(2.0, "meter")
        assert ureg.Quantity(2.0, "meter") >= ureg.Quantity(1.0, "meter")

    def test_compare_method(self, ureg: UnitRegistry) -> None:
        a = ureg.Quantity(1.0, "km")
        b = ureg.Quantity(500.0, "meter")
        assert a.compare(b, ">")
        assert not a.compare(b, "==")


class TestQuantityDimensionality:
    """Mirrors pint's dimensionality tests."""

    def test_dimensionality(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter")
        assert q.dimensionality == "[length]"
        assert not q.dimensionless

    def test_dimensionless(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter") / ureg.Quantity(1.0, "meter")
        assert q.dimensionless
        assert q.unitless

    def test_float_dimensionless(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.0, "meter") / ureg.Quantity(1.0, "meter")
        assert float(q) == 3.0

    def test_float_dimensioned_raises(self, ureg: UnitRegistry) -> None:
        with pytest.raises(DimensionalityError):
            float(ureg.Quantity(3.0, "meter"))

    def test_is_compatible_with(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter")
        assert q.is_compatible_with("foot")
        assert q.is_compatible_with("inch")
        assert not q.is_compatible_with("second")

    def test_check(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.0, "meter")
        assert q.check("[length]")
        assert not q.check("[time]")


class TestQuantityRepr:
    """Mirrors pint's repr/str tests."""

    def test_str(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.5, "meter")
        assert str(q) == "3.5 m"

    def test_repr(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.5, "meter")
        r = repr(q)
        assert "3.5" in r
        assert "m" in r

    def test_format_empty(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(3.5, "meter")
        assert format(q, "") == "3.5 m"


class TestQuantityPickle:
    """Mirrors pint's pickle tests."""

    def test_pickle(self) -> None:
        q = Quantity(5.0, "meter")
        q2 = pickle.loads(pickle.dumps(q))
        assert abs(q2.magnitude - 5.0) < 1e-10
        assert str(q2.units) == "m"

    def test_copy(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "meter")
        q2 = copy.copy(q)
        assert q.magnitude == q2.magnitude

    def test_deepcopy(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "meter")
        q2 = copy.deepcopy(q)
        assert q.magnitude == q2.magnitude


class TestOffsetUnits:
    """Mirrors pint's temperature tests."""

    def test_celsius_to_kelvin(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "degree_Celsius")
        result = q.to("kelvin")
        assert abs(result.magnitude - 373.15) < 1e-6

    def test_celsius_to_fahrenheit(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "degree_Celsius")
        result = q.to("degree_Fahrenheit")
        assert abs(result.magnitude - 212.0) < 1e-6

    def test_fahrenheit_to_celsius(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(32, "degree_Fahrenheit")
        result = q.to("degree_Celsius")
        assert abs(result.magnitude) < 1e-6

    def test_absolute_zero(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "kelvin")
        result = q.to("degree_Celsius")
        assert abs(result.magnitude - (-273.15)) < 1e-6


class TestTimedelta:
    """Mirrors pint's timedelta tests."""

    def test_to_timedelta(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(60, "second")
        td = q.to_timedelta()
        assert isinstance(td, datetime.timedelta)
        assert td.total_seconds() == 60.0

    def test_to_timedelta_minutes(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1.5, "minute")
        td = q.to_timedelta()
        assert abs(td.total_seconds() - 90.0) < 1e-6

    def test_to_timedelta_hours(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "hour")
        td = q.to_timedelta()
        assert abs(td.total_seconds() - 3600.0) < 1e-6


class TestUnitOperations:
    """Mirrors pint's unit operation tests."""

    def test_unit_from(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        q = u.from_(ureg.Quantity(1, "km"))
        assert abs(q.magnitude - 1000) < 1e-10

    def test_unit_m_from(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        assert abs(u.m_from(ureg.Quantity(1, "km")) - 1000) < 1e-10

    def test_unit_mul(self, ureg: UnitRegistry) -> None:
        u1 = ureg.Unit("meter")
        u2 = ureg.Unit("second")
        u3 = u1 * u2
        assert str(u3) != ""

    def test_unit_div(self, ureg: UnitRegistry) -> None:
        u1 = ureg.Unit("meter")
        u2 = ureg.Unit("second")
        u3 = u1 / u2
        assert str(u3) != ""

    def test_unit_pow(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        u2 = u**2
        assert str(u2) != ""

    def test_unit_eq(self, ureg: UnitRegistry) -> None:
        assert ureg.Unit("meter") == ureg.Unit("meter")

    def test_unit_is_compatible_with(self, ureg: UnitRegistry) -> None:
        u = ureg.Unit("meter")
        assert u.is_compatible_with("foot")
        assert not u.is_compatible_with("second")

    def test_unit_dimensionality(self, ureg: UnitRegistry) -> None:
        assert ureg.Unit("meter").dimensionality == "[length]"

    def test_unit_pickle(self) -> None:
        u = Unit("meter")
        u2 = pickle.loads(pickle.dumps(u))
        assert str(u) == str(u2)


class TestRegistryOperations:
    """Mirrors pint's registry tests."""

    def test_parse_expression(self, ureg: UnitRegistry) -> None:
        q = ureg.parse_expression("9.8 m/s^2")
        assert abs(q.magnitude - 9.8) < 1e-10

    def test_parse_units(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("m/s")
        assert str(u) != ""

    def test_getattr(self, ureg: UnitRegistry) -> None:
        q = ureg.meter
        assert q.magnitude == 1.0
        assert str(q.units) == "m"

    def test_contains(self, ureg: UnitRegistry) -> None:
        assert "meter" in ureg
        assert "foobar" not in ureg

    def test_iter(self, ureg: UnitRegistry) -> None:
        names = list(ureg)
        assert "meter" in names
        assert len(names) > 100

    def test_define(self, ureg: UnitRegistry) -> None:
        ureg.define("smoot = 1.7018 * meter")
        q = ureg.Quantity(1, "smoot").to("meter")
        assert abs(q.magnitude - 1.7018) < 1e-10

    def test_convert(self, ureg: UnitRegistry) -> None:
        assert abs(ureg.convert(1.0, "km", "meter") - 1000) < 1e-10

    def test_get_base_units(self, ureg: UnitRegistry) -> None:
        factor, _unit = ureg.get_base_units("kilometer")
        assert abs(factor - 1000) < 1e-10

    def test_get_name(self, ureg: UnitRegistry) -> None:
        assert ureg.get_name("m") == "meter"

    def test_get_symbol(self, ureg: UnitRegistry) -> None:
        assert ureg.get_symbol("meter") == "m"

    def test_is_compatible_with(self, ureg: UnitRegistry) -> None:
        assert ureg.is_compatible_with("meter", "foot")
        assert not ureg.is_compatible_with("meter", "second")

    def test_get_compatible_units(self, ureg: UnitRegistry) -> None:
        units = ureg.get_compatible_units("meter")
        assert len(units) > 5


class TestVariousUnits:
    """Mirrors pint's various unit conversion tests."""

    CONVERSIONS = [
        ("atmosphere", "pascal", 101325, 1),
        ("mile", "kilometer", 1.609344, 1e-4),
        ("gallon", "liter", 3.785412, 1e-3),
        ("horsepower", "watt", 745.7, 1.0),
        ("electron_volt", "joule", 1.602e-19, 1e-22),
        ("speed_of_light", "m/s", 299792458, 1),
        ("inch", "centimeter", 2.54, 1e-6),
        ("pound", "kilogram", 0.453592, 1e-3),
        ("foot", "meter", 0.3048, 1e-6),
        ("yard", "meter", 0.9144, 1e-6),
        ("acre", "m**2", 4046.86, 1),
        ("bar", "pascal", 1e5, 1),
        ("calorie", "joule", 4.184, 1e-3),
    ]

    @pytest.mark.parametrize(
        ("src", "dst", "expected", "tol"),
        CONVERSIONS,
        ids=[f"{s}->{d}" for s, d, _, _ in CONVERSIONS],
    )
    def test_conversion(
        self,
        ureg: UnitRegistry,
        src: str,
        dst: str,
        expected: float,
        tol: float,
    ) -> None:
        q = ureg.Quantity(1, src)
        result = q.to(dst)
        assert abs(result.magnitude - expected) < tol


class TestWrapsDecorator:
    """Mirrors pint's wraps tests."""

    def test_basic_wraps(self, ureg: UnitRegistry) -> None:
        @ureg.wraps("meter", ("meter", "meter"))
        def add_lengths(a: float, b: float) -> float:
            return a + b

        result = add_lengths(ureg.Quantity(1, "km"), ureg.Quantity(500, "meter"))
        assert abs(result.magnitude - 1500) < 1e-10

    def test_check_decorator(self, ureg: UnitRegistry) -> None:
        @ureg.check("[length]", "[time]")
        def speed(d: Quantity, t: Quantity) -> Quantity:
            return d / t

        result = speed(ureg.Quantity(100, "meter"), ureg.Quantity(10, "second"))
        assert abs(result.magnitude - 10) < 1e-10

    def test_check_fails(self, ureg: UnitRegistry) -> None:
        @ureg.check("[length]", "[time]")
        def speed(d: Quantity, t: Quantity) -> Quantity:
            return d / t

        with pytest.raises(DimensionalityError):
            speed(
                ureg.Quantity(100, "meter"),
                ureg.Quantity(10, "kilogram"),
            )


class TestMeasurement:
    """Mirrors pint's measurement tests."""

    def test_plus_minus(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(5.0, "meter")
        m = q.plus_minus(0.1)
        assert isinstance(m, Measurement)
        assert m.magnitude == 5.0
        assert abs(m.rel - 0.02) < 1e-10

    def test_measurement_constructor(self, ureg: UnitRegistry) -> None:
        m = ureg.Measurement(5.0, 0.1, "meter")
        assert isinstance(m, Measurement)
        assert m.magnitude == 5.0

    def test_measurement_str(self, ureg: UnitRegistry) -> None:
        m = Measurement(ureg.Quantity(5.0, "meter"), 0.1)
        s = str(m)
        assert "5.0" in s
        assert "0.1" in s



"""Ported from pint's test_errors.py - comprehensive exception tests."""

from __future__ import annotations

import pickle

import pytest
from pintrs import (
    DefinitionSyntaxError,
    DimensionalityError,
    OffsetUnitCalculusError,
    PintError,
    RedefinitionError,
    UndefinedUnitError,
    UnitRegistry,
)


class TestExceptionHierarchy:
    """All pint exceptions inherit from PintError which inherits from ValueError."""

    def test_pint_error_is_value_error(self) -> None:
        assert issubclass(PintError, ValueError)

    def test_dimensionality_error_is_pint_error(self) -> None:
        assert issubclass(DimensionalityError, PintError)

    def test_undefined_unit_error_is_pint_error(self) -> None:
        assert issubclass(UndefinedUnitError, PintError)

    def test_offset_unit_calculus_error_is_pint_error(self) -> None:
        assert issubclass(OffsetUnitCalculusError, PintError)

    def test_definition_syntax_error_is_pint_error(self) -> None:
        assert issubclass(DefinitionSyntaxError, PintError)

    def test_redefinition_error_is_pint_error(self) -> None:
        assert issubclass(RedefinitionError, PintError)

    def test_dimensionality_error_is_value_error(self) -> None:
        assert issubclass(DimensionalityError, ValueError)

    def test_undefined_unit_error_is_value_error(self) -> None:
        assert issubclass(UndefinedUnitError, ValueError)


class TestExceptionRaising:
    """Exceptions are raised in the correct situations."""

    def test_incompatible_conversion_raises_dimensionality_error(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises(DimensionalityError):
            ureg.Quantity(1, "meter").to("second")

    def test_unknown_unit_raises_undefined_unit_error(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises(UndefinedUnitError):
            ureg.Quantity(1, "gibberish_unit")

    def test_redefinition_raises(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises((RedefinitionError, DefinitionSyntaxError)):
            ureg.define("meter = [length]")

    def test_bad_definition_raises(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises((DefinitionSyntaxError, ValueError)):
            ureg.define("= = = = =")

    def test_catch_as_pint_error(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises(PintError):
            ureg.Quantity(1, "meter").to("second")

    def test_catch_as_value_error(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises(ValueError):
            ureg.Quantity(1, "meter").to("second")


class TestExceptionPickle:
    """Exceptions can be pickled and unpickled."""

    def test_pickle_dimensionality_error(self) -> None:
        ureg = UnitRegistry()
        try:
            ureg.Quantity(1, "meter").to("second")
        except DimensionalityError as e:
            restored = pickle.loads(pickle.dumps(e))
            assert isinstance(restored, DimensionalityError)

    def test_pickle_undefined_unit_error(self) -> None:
        ureg = UnitRegistry()
        try:
            ureg.Quantity(1, "gibberish_unit_xyz")
        except UndefinedUnitError as e:
            restored = pickle.loads(pickle.dumps(e))
            assert isinstance(restored, UndefinedUnitError)


class TestExceptionMessages:
    """Exception messages are informative."""

    def test_dimensionality_error_mentions_units(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises(DimensionalityError, match=r"meter|second|length|time"):
            ureg.Quantity(1, "meter").to("second")

    def test_undefined_unit_error_mentions_unit_name(self) -> None:
        ureg = UnitRegistry()
        with pytest.raises(UndefinedUnitError, match="nonexistent_xyz"):
            ureg.Quantity(1, "nonexistent_xyz")



"""Tests for pandas ExtensionArray integration."""

from __future__ import annotations

import pytest

pd = pytest.importorskip("pandas")
np = pytest.importorskip("numpy")

from pintrs import UnitRegistry
from pintrs.pandas_support import PintArray, PintDtype


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestPintDtype:
    def test_name(self) -> None:
        dtype = PintDtype("meter")
        assert dtype.name == "pint[meter]"

    def test_from_string(self) -> None:
        dtype = PintDtype.construct_from_string("pint[meter]")
        assert dtype.units == "meter"

    def test_invalid_string(self) -> None:
        with pytest.raises(TypeError):
            PintDtype.construct_from_string("int64")

    def test_eq(self) -> None:
        assert PintDtype("meter") == PintDtype("meter")
        assert PintDtype("meter") != PintDtype("second")

    def test_repr(self) -> None:
        dtype = PintDtype("meter")
        assert "meter" in repr(dtype)


class TestPintArray:
    def test_create(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        arr = PintArray([1.0, 2.0, 3.0], dtype=dtype)
        assert len(arr) == 3

    def test_getitem_scalar(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        arr = PintArray([1.0, 2.0, 3.0], dtype=dtype)
        item = arr[0]
        assert item.magnitude == 1.0

    def test_getitem_slice(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        arr = PintArray([1.0, 2.0, 3.0], dtype=dtype)
        sliced = arr[0:2]
        assert isinstance(sliced, PintArray)
        assert len(sliced) == 2

    def test_setitem(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        arr = PintArray([1.0, 2.0, 3.0], dtype=dtype)
        arr[1] = 5.0
        assert arr[1].magnitude == 5.0

    def test_isna(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        arr = PintArray([1.0, float("nan"), 3.0], dtype=dtype)
        na_mask = arr.isna()
        assert not na_mask[0]
        assert na_mask[1]
        assert not na_mask[2]

    def test_to_conversion(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("kilometer", ureg)
        arr = PintArray([1.0, 2.0], dtype=dtype)
        result = arr.to("meter")
        assert abs(result[0].magnitude - 1000.0) < 1e-10
        assert result.units == "meter"

    def test_copy(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        arr = PintArray([1.0, 2.0], dtype=dtype)
        arr2 = arr.copy()
        arr2[0] = 99.0
        assert arr[0].magnitude == 1.0

    def test_concat(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        a = PintArray([1.0, 2.0], dtype=dtype)
        b = PintArray([3.0, 4.0], dtype=dtype)
        result = PintArray._concat_same_type([a, b])
        assert len(result) == 4


class TestPandasIntegration:
    def test_series_creation(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        s = pd.Series(PintArray([1.0, 2.0, 3.0], dtype=dtype))
        assert len(s) == 3

    def test_dataframe_column(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        df = pd.DataFrame({"distance": PintArray([1.0, 2.0], dtype=dtype)})
        assert len(df) == 2
        assert df["distance"].dtype == dtype

    def test_series_sum(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        s = pd.Series(PintArray([1.0, 2.0, 3.0], dtype=dtype))
        total = s.sum()
        assert abs(total.magnitude - 6.0) < 1e-10

    def test_series_mean(self, ureg: UnitRegistry) -> None:
        dtype = PintDtype("meter", ureg)
        s = pd.Series(PintArray([1.0, 2.0, 3.0], dtype=dtype))
        avg = s.mean()
        assert abs(avg.magnitude - 2.0) < 1e-10
